/*
 * Copyright (C) 2010-2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "EventDispatcher.h"

#include "Decoder.h"
#include "EventDispatcherMessages.h"
#include "HandleMessage.h"
#include "WebCoreArgumentCoders.h"
#if ENABLE(MAC_GESTURE_EVENTS)
#include "WebGestureEvent.h"
#endif
#if ENABLE(IOS_TOUCH_EVENTS)
#include "WebTouchEvent.h"
#endif
#include "WebWheelEvent.h"
#include <WebCore/PageIdentifier.h>

namespace Messages {

namespace EventDispatcher {

#if ENABLE(IOS_TOUCH_EVENTS)

void TouchEvent::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    Optional<bool> handled;
    decoder >> handled;
    if (!handled) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*handled));
}

void TouchEvent::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<bool>::create());
}

void TouchEvent::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, bool handled)
{
    *encoder << handled;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

} // namespace EventDispatcher

} // namespace Messages

namespace WebKit {

void EventDispatcher::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    auto protectedThis = makeRef(*this);
    if (decoder.messageName() == Messages::EventDispatcher::WheelEvent::name()) {
        IPC::handleMessage<Messages::EventDispatcher::WheelEvent>(decoder, this, &EventDispatcher::wheelEvent);
        return;
    }
#if ENABLE(IOS_TOUCH_EVENTS)
    if (decoder.messageName() == Messages::EventDispatcher::TouchEvent::name()) {
        IPC::handleMessageAsync<Messages::EventDispatcher::TouchEvent>(connection, decoder, this, &EventDispatcher::touchEvent);
        return;
    }
#endif
#if ENABLE(IOS_TOUCH_EVENTS)
    if (decoder.messageName() == Messages::EventDispatcher::TouchEventWithoutCallback::name()) {
        IPC::handleMessage<Messages::EventDispatcher::TouchEventWithoutCallback>(decoder, this, &EventDispatcher::touchEventWithoutCallback);
        return;
    }
#endif
#if ENABLE(MAC_GESTURE_EVENTS)
    if (decoder.messageName() == Messages::EventDispatcher::GestureEvent::name()) {
        IPC::handleMessage<Messages::EventDispatcher::GestureEvent>(decoder, this, &EventDispatcher::gestureEvent);
        return;
    }
#endif
#if ENABLE(WEBPROCESS_WINDOWSERVER_BLOCKING)
    if (decoder.messageName() == Messages::EventDispatcher::DisplayWasRefreshed::name()) {
        IPC::handleMessage<Messages::EventDispatcher::DisplayWasRefreshed>(decoder, this, &EventDispatcher::displayWasRefreshed);
        return;
    }
#endif
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit
