/*
 * Copyright (C) 2010-2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#if ENABLE(SERVICE_WORKER)
#include "WebSWServerConnection.h"

#include "ArgumentCoders.h"
#include "Decoder.h"
#include "HandleMessage.h"
#include "WebCoreArgumentCoders.h"
#include "WebSWServerConnectionMessages.h"
#include <WebCore/ExceptionData.h>
#include <WebCore/MessageWithMessagePorts.h>
#include <WebCore/SecurityOriginData.h>
#include <WebCore/ServiceWorkerClientData.h>
#include <WebCore/ServiceWorkerClientIdentifier.h>
#include <WebCore/ServiceWorkerFetchResult.h>
#include <WebCore/ServiceWorkerIdentifier.h>
#include <WebCore/ServiceWorkerJobData.h>
#include <WebCore/ServiceWorkerRegistrationKey.h>
#include <WebCore/ServiceWorkerTypes.h>
#include <wtf/Expected.h>
#include <wtf/Optional.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace WebSWServerConnection {

void ScheduleUnregisterJobInServer::callReply(IPC::Decoder& decoder, CompletionHandler<void(Expected<bool, WebCore::ExceptionData>&&)>&& completionHandler)
{
    Optional<Expected<bool, WebCore::ExceptionData>> result;
    decoder >> result;
    if (!result) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*result));
}

void ScheduleUnregisterJobInServer::cancelReply(CompletionHandler<void(Expected<bool, WebCore::ExceptionData>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Expected<bool, WebCore::ExceptionData>>::create());
}

void ScheduleUnregisterJobInServer::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const Expected<bool, WebCore::ExceptionData>& result)
{
    *encoder << result;
    connection.sendSyncReply(WTFMove(encoder));
}

void TerminateWorkerFromClient::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void TerminateWorkerFromClient::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void TerminateWorkerFromClient::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

void WhenServiceWorkerIsTerminatedForTesting::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void WhenServiceWorkerIsTerminatedForTesting::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void WhenServiceWorkerIsTerminatedForTesting::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

void StoreRegistrationsOnDisk::callReply(IPC::Decoder& decoder, CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void StoreRegistrationsOnDisk::cancelReply(CompletionHandler<void()>&& completionHandler)
{
    completionHandler();
}

void StoreRegistrationsOnDisk::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection)
{
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace WebSWServerConnection

} // namespace Messages

namespace WebKit {

void WebSWServerConnection::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    if (decoder.messageName() == Messages::WebSWServerConnection::ScheduleJobInServer::name()) {
        IPC::handleMessage<Messages::WebSWServerConnection::ScheduleJobInServer>(decoder, this, &WebSWServerConnection::scheduleJobInServer);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerConnection::ScheduleUnregisterJobInServer::name()) {
        IPC::handleMessageAsync<Messages::WebSWServerConnection::ScheduleUnregisterJobInServer>(connection, decoder, this, &WebSWServerConnection::scheduleUnregisterJobInServer);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerConnection::FinishFetchingScriptInServer::name()) {
        IPC::handleMessage<Messages::WebSWServerConnection::FinishFetchingScriptInServer>(decoder, this, &WebSWServerConnection::finishFetchingScriptInServer);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerConnection::AddServiceWorkerRegistrationInServer::name()) {
        IPC::handleMessage<Messages::WebSWServerConnection::AddServiceWorkerRegistrationInServer>(decoder, this, &WebSWServerConnection::addServiceWorkerRegistrationInServer);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerConnection::RemoveServiceWorkerRegistrationInServer::name()) {
        IPC::handleMessage<Messages::WebSWServerConnection::RemoveServiceWorkerRegistrationInServer>(decoder, this, &WebSWServerConnection::removeServiceWorkerRegistrationInServer);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerConnection::PostMessageToServiceWorker::name()) {
        IPC::handleMessage<Messages::WebSWServerConnection::PostMessageToServiceWorker>(decoder, this, &WebSWServerConnection::postMessageToServiceWorker);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerConnection::DidResolveRegistrationPromise::name()) {
        IPC::handleMessage<Messages::WebSWServerConnection::DidResolveRegistrationPromise>(decoder, this, &WebSWServerConnection::didResolveRegistrationPromise);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerConnection::MatchRegistration::name()) {
        IPC::handleMessage<Messages::WebSWServerConnection::MatchRegistration>(decoder, this, &WebSWServerConnection::matchRegistration);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerConnection::WhenRegistrationReady::name()) {
        IPC::handleMessage<Messages::WebSWServerConnection::WhenRegistrationReady>(decoder, this, &WebSWServerConnection::whenRegistrationReady);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerConnection::GetRegistrations::name()) {
        IPC::handleMessage<Messages::WebSWServerConnection::GetRegistrations>(decoder, this, &WebSWServerConnection::getRegistrations);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerConnection::RegisterServiceWorkerClient::name()) {
        IPC::handleMessage<Messages::WebSWServerConnection::RegisterServiceWorkerClient>(decoder, this, &WebSWServerConnection::registerServiceWorkerClient);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerConnection::UnregisterServiceWorkerClient::name()) {
        IPC::handleMessage<Messages::WebSWServerConnection::UnregisterServiceWorkerClient>(decoder, this, &WebSWServerConnection::unregisterServiceWorkerClient);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerConnection::TerminateWorkerFromClient::name()) {
        IPC::handleMessageAsync<Messages::WebSWServerConnection::TerminateWorkerFromClient>(connection, decoder, this, &WebSWServerConnection::terminateWorkerFromClient);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerConnection::WhenServiceWorkerIsTerminatedForTesting::name()) {
        IPC::handleMessageAsync<Messages::WebSWServerConnection::WhenServiceWorkerIsTerminatedForTesting>(connection, decoder, this, &WebSWServerConnection::whenServiceWorkerIsTerminatedForTesting);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerConnection::SetThrottleState::name()) {
        IPC::handleMessage<Messages::WebSWServerConnection::SetThrottleState>(decoder, this, &WebSWServerConnection::setThrottleState);
        return;
    }
    if (decoder.messageName() == Messages::WebSWServerConnection::StoreRegistrationsOnDisk::name()) {
        IPC::handleMessageAsync<Messages::WebSWServerConnection::StoreRegistrationsOnDisk>(connection, decoder, this, &WebSWServerConnection::storeRegistrationsOnDisk);
        return;
    }
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit

#endif // ENABLE(SERVICE_WORKER)
