/*
 * Copyright (C) 2010-2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ArgumentCoders.h"
#include "Connection.h"
#include "ContentWorldShared.h"
#include "DataReference.h"
#include "DrawingAreaInfo.h"
#include "EditorState.h"
#include "GamepadData.h"
#include "GeolocationIdentifier.h"
#if PLATFORM(IOS_FAMILY)
#include "GestureTypes.h"
#endif
#include "MessageNames.h"
#include "PDFPluginIdentifier.h"
#include "SandboxExtension.h"
#include "SessionState.h"
#include "ShareableBitmap.h"
#include "SharedBufferDataReference.h"
#include "SharedMemory.h"
#include "TransactionID.h"
#include "WebEvent.h"
#include "WebFindOptions.h"
#include "WebPageMessagesReplies.h"
#include "WebPageProxyIdentifier.h"
#include "WebsitePoliciesData.h"
#include <JavaScriptCore/InspectorFrontendChannel.h>
#include <WebCore/ActivityState.h>
#include <WebCore/ApplicationManifest.h>
#include <WebCore/Color.h>
#include <WebCore/CompositionHighlight.h>
#include <WebCore/CompositionUnderline.h>
#include <WebCore/DictationAlternative.h>
#include <WebCore/DragActions.h>
#include <WebCore/DragData.h>
#include <WebCore/ElementContext.h>
#include <WebCore/FloatRect.h>
#include <WebCore/FrameIdentifier.h>
#include <WebCore/IntRect.h>
#include <WebCore/IntSize.h>
#include <WebCore/JSDOMExceptionHandling.h>
#include <WebCore/LayoutMilestone.h>
#include <WebCore/MediaProducer.h>
#include <WebCore/PageOverlay.h>
#include <WebCore/PlatformEvent.h>
#include <WebCore/PlaybackTargetClientContextIdentifier.h>
#include <WebCore/PointerID.h>
#include <WebCore/RectEdges.h>
#include <WebCore/RegistrableDomain.h>
#include <WebCore/SecurityPolicyViolationEvent.h>
#include <WebCore/TextCheckerClient.h>
#include <WebCore/TextIndicator.h>
#include <WebCore/TextManipulationController.h>
#include <WebCore/ViewportArguments.h>
#include <utility>
#include <wtf/Forward.h>
#include <wtf/MachSendRight.h>
#include <wtf/OptionSet.h>
#include <wtf/Optional.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace IPC {
class FormDataReference;
class SharedBufferDataReference;
}

namespace JSC {
enum class MessageLevel : uint8_t;
enum class MessageSource : uint8_t;
}

namespace WebCore {
class CaptureDevice;
class Color;
class DragData;
class FloatPoint;
class FloatRect;
class FloatSize;
class FontAttributeChanges;
class FontChanges;
class IntPoint;
class IntRect;
class IntSize;
class MediaPlaybackTargetContext;
class ResourceError;
class ResourceRequest;
class ResourceResponse;
class SelectionData;
enum class CreateNewGroupForHighlight : bool;
enum class EventMakesGamepadsVisible : bool;
enum class ShouldTreatAsContinuingLoad : bool;
enum class FrameLoadType : uint8_t;
enum class SelectionDirection : uint8_t;
enum class TextGranularity : uint8_t;
enum class WritingDirection : uint8_t;
struct AttributedString;
struct BackForwardItemIdentifier;
struct ElementContext;
struct FontAttributes;
struct GlobalFrameIdentifier;
struct GlobalWindowIdentifier;
struct RunJavaScriptParameters;
struct SystemPreviewInfo;
struct TextCheckingResult;
}

namespace WebKit {
class CallbackID;
class UserData;
class WebContextMenuItemData;
class WebKeyboardEvent;
class WebMouseEvent;
class WebTouchEvent;
class WebWheelEvent;
enum class ContentAsStringIncludesChildFrames : bool;
enum class RespectSelectionAnchor : bool;
enum class DragControllerAction : uint8_t;
enum class GestureRecognizerState : uint8_t;
enum class GestureType : uint8_t;
enum class LayerHostingMode : uint8_t;
enum class MediaPlaybackState : uint8_t;
enum class SelectionTouch : uint8_t;
enum class SyntheticEditingCommandType : uint8_t;
struct BackForwardListItemState;
struct DataDetectionResult;
struct DocumentEditingContext;
struct DocumentEditingContextRequest;
struct EditingRange;
struct FocusedElementInformation;
struct FontInfo;
struct FrameInfoData;
struct FrameTreeNodeData;
struct InsertTextOptions;
struct InteractionInformationAtPosition;
struct InteractionInformationRequest;
struct LoadParameters;
struct PolicyDecision;
struct PrintInfo;
struct UserMessage;
struct WebAutocorrectionData;
struct WebPreferencesStore;
struct WebsitePoliciesData;
}

namespace Messages {
namespace WebPage {

static inline IPC::ReceiverName messageReceiverName()
{
    return IPC::ReceiverName::WebPage;
}

class SetInitialFocus {
public:
    using Arguments = std::tuple<bool, bool, const WebKit::WebKeyboardEvent&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetInitialFocus; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_SetInitialFocusReply; }
    using AsyncReply = SetInitialFocusAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    SetInitialFocus(bool forward, bool isKeyboardEventValid, const WebKit::WebKeyboardEvent& event)
        : m_arguments(forward, isKeyboardEventValid, event)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetActivityState {
public:
    using Arguments = std::tuple<const OptionSet<WebCore::ActivityState::Flag>&, const WebKit::ActivityStateChangeID&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetActivityState; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_SetActivityStateReply; }
    using AsyncReply = SetActivityStateAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    SetActivityState(const OptionSet<WebCore::ActivityState::Flag>& activityState, const WebKit::ActivityStateChangeID& activityStateChangeID)
        : m_arguments(activityState, activityStateChangeID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetLayerHostingMode {
public:
    using Arguments = std::tuple<WebKit::LayerHostingMode>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetLayerHostingMode; }
    static const bool isSync = false;

    explicit SetLayerHostingMode(WebKit::LayerHostingMode layerHostingMode)
        : m_arguments(layerHostingMode)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetBackgroundColor {
public:
    using Arguments = std::tuple<const Optional<WebCore::Color>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetBackgroundColor; }
    static const bool isSync = false;

    explicit SetBackgroundColor(const Optional<WebCore::Color>& color)
        : m_arguments(color)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class AddConsoleMessage {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&, JSC::MessageSource, JSC::MessageLevel, const String&, uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_AddConsoleMessage; }
    static const bool isSync = false;

    AddConsoleMessage(const WebCore::FrameIdentifier& frameID, JSC::MessageSource messageSource, JSC::MessageLevel messageLevel, const String& message, uint64_t requestID)
        : m_arguments(frameID, messageSource, messageLevel, message, requestID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SendCSPViolationReport {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&, const URL&, const IPC::FormDataReference&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SendCSPViolationReport; }
    static const bool isSync = false;

    SendCSPViolationReport(const WebCore::FrameIdentifier& frameID, const URL& reportURL, const IPC::FormDataReference& reportData)
        : m_arguments(frameID, reportURL, reportData)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class EnqueueSecurityPolicyViolationEvent {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&, const WebCore::SecurityPolicyViolationEvent::Init&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_EnqueueSecurityPolicyViolationEvent; }
    static const bool isSync = false;

    EnqueueSecurityPolicyViolationEvent(const WebCore::FrameIdentifier& frameID, const WebCore::SecurityPolicyViolationEvent::Init& eventInit)
        : m_arguments(frameID, eventInit)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class TestProcessIncomingSyncMessagesWhenWaitingForSyncReply {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_TestProcessIncomingSyncMessagesWhenWaitingForSyncReply; }
    static const bool isSync = true;

    using DelayedReply = TestProcessIncomingSyncMessagesWhenWaitingForSyncReplyDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool handled);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(COCOA)
class SetTopContentInsetFenced {
public:
    using Arguments = std::tuple<float, const MachSendRight&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetTopContentInsetFenced; }
    static const bool isSync = false;

    SetTopContentInsetFenced(float contentInset, const MachSendRight& machSendRight)
        : m_arguments(contentInset, machSendRight)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class SetTopContentInset {
public:
    using Arguments = std::tuple<float>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetTopContentInset; }
    static const bool isSync = false;

    explicit SetTopContentInset(float contentInset)
        : m_arguments(contentInset)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetUnderlayColor {
public:
    using Arguments = std::tuple<const WebCore::Color&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetUnderlayColor; }
    static const bool isSync = false;

    explicit SetUnderlayColor(const WebCore::Color& color)
        : m_arguments(color)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ViewWillStartLiveResize {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ViewWillStartLiveResize; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ViewWillEndLiveResize {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ViewWillEndLiveResize; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ExecuteEditCommandWithCallback {
public:
    using Arguments = std::tuple<const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ExecuteEditCommandWithCallback; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_ExecuteEditCommandWithCallbackReply; }
    using AsyncReply = ExecuteEditCommandWithCallbackAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    ExecuteEditCommandWithCallback(const String& name, const String& argument)
        : m_arguments(name, argument)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class KeyEvent {
public:
    using Arguments = std::tuple<const WebKit::WebKeyboardEvent&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_KeyEvent; }
    static const bool isSync = false;

    explicit KeyEvent(const WebKit::WebKeyboardEvent& event)
        : m_arguments(event)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class MouseEvent {
public:
    using Arguments = std::tuple<const WebKit::WebMouseEvent&, const Optional<WebKit::SandboxExtension::HandleArray>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_MouseEvent; }
    static const bool isSync = false;

    MouseEvent(const WebKit::WebMouseEvent& event, const Optional<WebKit::SandboxExtension::HandleArray>& sandboxExtensions)
        : m_arguments(event, sandboxExtensions)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(IOS_FAMILY)
class SetViewportConfigurationViewLayoutSize {
public:
    using Arguments = std::tuple<const WebCore::FloatSize&, double, double>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetViewportConfigurationViewLayoutSize; }
    static const bool isSync = false;

    SetViewportConfigurationViewLayoutSize(const WebCore::FloatSize& size, double scaleFactor, double minimumEffectiveDeviceWidth)
        : m_arguments(size, scaleFactor, minimumEffectiveDeviceWidth)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class SetMaximumUnobscuredSize {
public:
    using Arguments = std::tuple<const WebCore::FloatSize&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetMaximumUnobscuredSize; }
    static const bool isSync = false;

    explicit SetMaximumUnobscuredSize(const WebCore::FloatSize& size)
        : m_arguments(size)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class SetDeviceOrientation {
public:
    using Arguments = std::tuple<int32_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetDeviceOrientation; }
    static const bool isSync = false;

    explicit SetDeviceOrientation(int32_t deviceOrientation)
        : m_arguments(deviceOrientation)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class SetOverrideViewportArguments {
public:
    using Arguments = std::tuple<const Optional<WebCore::ViewportArguments>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetOverrideViewportArguments; }
    static const bool isSync = false;

    explicit SetOverrideViewportArguments(const Optional<WebCore::ViewportArguments>& arguments)
        : m_arguments(arguments)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class DynamicViewportSizeUpdate {
public:
    using Arguments = std::tuple<const WebCore::FloatSize&, const WebCore::FloatSize&, const WebCore::FloatRect&, const WebCore::FloatRect&, const WebCore::FloatRect&, const WebCore::RectEdges<float>&, double, int32_t, double, uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DynamicViewportSizeUpdate; }
    static const bool isSync = false;

    DynamicViewportSizeUpdate(const WebCore::FloatSize& viewLayoutSize, const WebCore::FloatSize& maximumUnobscuredSize, const WebCore::FloatRect& targetExposedContentRect, const WebCore::FloatRect& targetUnobscuredRect, const WebCore::FloatRect& targetUnobscuredRectInScrollViewCoordinates, const WebCore::RectEdges<float>& targetUnobscuredSafeAreaInsets, double scale, int32_t deviceOrientation, double minimumEffectiveDeviceWidth, uint64_t dynamicViewportSizeUpdateID)
        : m_arguments(viewLayoutSize, maximumUnobscuredSize, targetExposedContentRect, targetUnobscuredRect, targetUnobscuredRectInScrollViewCoordinates, targetUnobscuredSafeAreaInsets, scale, deviceOrientation, minimumEffectiveDeviceWidth, dynamicViewportSizeUpdateID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class SetScreenIsBeingCaptured {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetScreenIsBeingCaptured; }
    static const bool isSync = false;

    explicit SetScreenIsBeingCaptured(bool captured)
        : m_arguments(captured)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class AttemptSyntheticClick {
public:
    using Arguments = std::tuple<const WebCore::IntPoint&, const OptionSet<WebKit::WebEvent::Modifier>&, const WebKit::TransactionID&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_AttemptSyntheticClick; }
    static const bool isSync = false;

    AttemptSyntheticClick(const WebCore::IntPoint& point, const OptionSet<WebKit::WebEvent::Modifier>& modifiers, const WebKit::TransactionID& lastLayerTreeTransactionId)
        : m_arguments(point, modifiers, lastLayerTreeTransactionId)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class PotentialTapAtPosition {
public:
    using Arguments = std::tuple<uint64_t, const WebCore::FloatPoint&, bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_PotentialTapAtPosition; }
    static const bool isSync = false;

    PotentialTapAtPosition(uint64_t requestID, const WebCore::FloatPoint& point, bool shouldRequestMagnificationInformation)
        : m_arguments(requestID, point, shouldRequestMagnificationInformation)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class CommitPotentialTap {
public:
    using Arguments = std::tuple<const OptionSet<WebKit::WebEvent::Modifier>&, const WebKit::TransactionID&, const WebCore::PointerID&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_CommitPotentialTap; }
    static const bool isSync = false;

    CommitPotentialTap(const OptionSet<WebKit::WebEvent::Modifier>& modifiers, const WebKit::TransactionID& lastLayerTreeTransactionId, const WebCore::PointerID& pointerId)
        : m_arguments(modifiers, lastLayerTreeTransactionId, pointerId)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class CancelPotentialTap {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_CancelPotentialTap; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class TapHighlightAtPosition {
public:
    using Arguments = std::tuple<uint64_t, const WebCore::FloatPoint&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_TapHighlightAtPosition; }
    static const bool isSync = false;

    TapHighlightAtPosition(uint64_t requestID, const WebCore::FloatPoint& point)
        : m_arguments(requestID, point)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class DidRecognizeLongPress {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidRecognizeLongPress; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class HandleDoubleTapForDoubleClickAtPoint {
public:
    using Arguments = std::tuple<const WebCore::IntPoint&, const OptionSet<WebKit::WebEvent::Modifier>&, const WebKit::TransactionID&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_HandleDoubleTapForDoubleClickAtPoint; }
    static const bool isSync = false;

    HandleDoubleTapForDoubleClickAtPoint(const WebCore::IntPoint& point, const OptionSet<WebKit::WebEvent::Modifier>& modifiers, const WebKit::TransactionID& lastLayerTreeTransactionId)
        : m_arguments(point, modifiers, lastLayerTreeTransactionId)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class InspectorNodeSearchMovedToPosition {
public:
    using Arguments = std::tuple<const WebCore::FloatPoint&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_InspectorNodeSearchMovedToPosition; }
    static const bool isSync = false;

    explicit InspectorNodeSearchMovedToPosition(const WebCore::FloatPoint& point)
        : m_arguments(point)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class InspectorNodeSearchEndedAtPosition {
public:
    using Arguments = std::tuple<const WebCore::FloatPoint&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_InspectorNodeSearchEndedAtPosition; }
    static const bool isSync = false;

    explicit InspectorNodeSearchEndedAtPosition(const WebCore::FloatPoint& point)
        : m_arguments(point)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class BlurFocusedElement {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_BlurFocusedElement; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class SelectWithGesture {
public:
    using Arguments = std::tuple<const WebCore::IntPoint&, WebKit::GestureType, WebKit::GestureRecognizerState, bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SelectWithGesture; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebCore::IntPoint&&, WebKit::GestureType&&, WebKit::GestureRecognizerState&&, OptionSet<WebKit::SelectionFlags>&&)>&&);
    static void cancelReply(CompletionHandler<void(WebCore::IntPoint&&, WebKit::GestureType&&, WebKit::GestureRecognizerState&&, OptionSet<WebKit::SelectionFlags>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_SelectWithGestureReply; }
    using AsyncReply = SelectWithGestureAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebCore::IntPoint& point, WebKit::GestureType gestureType, WebKit::GestureRecognizerState gestureState, const OptionSet<WebKit::SelectionFlags>& flags);
    using Reply = std::tuple<WebCore::IntPoint&, WebKit::GestureType&, WebKit::GestureRecognizerState&, OptionSet<WebKit::SelectionFlags>&>;
    using ReplyArguments = std::tuple<WebCore::IntPoint, WebKit::GestureType, WebKit::GestureRecognizerState, OptionSet<WebKit::SelectionFlags>>;
    SelectWithGesture(const WebCore::IntPoint& point, WebKit::GestureType gestureType, WebKit::GestureRecognizerState gestureState, bool isInteractingWithFocusedElement)
        : m_arguments(point, gestureType, gestureState, isInteractingWithFocusedElement)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class UpdateSelectionWithTouches {
public:
    using Arguments = std::tuple<const WebCore::IntPoint&, WebKit::SelectionTouch, bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_UpdateSelectionWithTouches; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebCore::IntPoint&&, WebKit::SelectionTouch&&, OptionSet<WebKit::SelectionFlags>&&)>&&);
    static void cancelReply(CompletionHandler<void(WebCore::IntPoint&&, WebKit::SelectionTouch&&, OptionSet<WebKit::SelectionFlags>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_UpdateSelectionWithTouchesReply; }
    using AsyncReply = UpdateSelectionWithTouchesAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebCore::IntPoint& point, WebKit::SelectionTouch touch, const OptionSet<WebKit::SelectionFlags>& selectionFlags);
    using Reply = std::tuple<WebCore::IntPoint&, WebKit::SelectionTouch&, OptionSet<WebKit::SelectionFlags>&>;
    using ReplyArguments = std::tuple<WebCore::IntPoint, WebKit::SelectionTouch, OptionSet<WebKit::SelectionFlags>>;
    UpdateSelectionWithTouches(const WebCore::IntPoint& point, WebKit::SelectionTouch touches, bool baseIsStart)
        : m_arguments(point, touches, baseIsStart)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class SelectWithTwoTouches {
public:
    using Arguments = std::tuple<const WebCore::IntPoint&, const WebCore::IntPoint&, WebKit::GestureType, WebKit::GestureRecognizerState>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SelectWithTwoTouches; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebCore::IntPoint&&, WebKit::GestureType&&, WebKit::GestureRecognizerState&&, OptionSet<WebKit::SelectionFlags>&&)>&&);
    static void cancelReply(CompletionHandler<void(WebCore::IntPoint&&, WebKit::GestureType&&, WebKit::GestureRecognizerState&&, OptionSet<WebKit::SelectionFlags>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_SelectWithTwoTouchesReply; }
    using AsyncReply = SelectWithTwoTouchesAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebCore::IntPoint& point, WebKit::GestureType gestureType, WebKit::GestureRecognizerState gestureState, const OptionSet<WebKit::SelectionFlags>& selectionFlags);
    using Reply = std::tuple<WebCore::IntPoint&, WebKit::GestureType&, WebKit::GestureRecognizerState&, OptionSet<WebKit::SelectionFlags>&>;
    using ReplyArguments = std::tuple<WebCore::IntPoint, WebKit::GestureType, WebKit::GestureRecognizerState, OptionSet<WebKit::SelectionFlags>>;
    SelectWithTwoTouches(const WebCore::IntPoint& from, const WebCore::IntPoint& to, WebKit::GestureType gestureType, WebKit::GestureRecognizerState gestureState)
        : m_arguments(from, to, gestureType, gestureState)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class ExtendSelection {
public:
    using Arguments = std::tuple<WebCore::TextGranularity>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ExtendSelection; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_ExtendSelectionReply; }
    using AsyncReply = ExtendSelectionAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    explicit ExtendSelection(WebCore::TextGranularity granularity)
        : m_arguments(granularity)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class SelectWordBackward {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SelectWordBackward; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class MoveSelectionByOffset {
public:
    using Arguments = std::tuple<int32_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_MoveSelectionByOffset; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_MoveSelectionByOffsetReply; }
    using AsyncReply = MoveSelectionByOffsetAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    explicit MoveSelectionByOffset(int32_t offset)
        : m_arguments(offset)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class SelectTextWithGranularityAtPoint {
public:
    using Arguments = std::tuple<const WebCore::IntPoint&, WebCore::TextGranularity, bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SelectTextWithGranularityAtPoint; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_SelectTextWithGranularityAtPointReply; }
    using AsyncReply = SelectTextWithGranularityAtPointAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    SelectTextWithGranularityAtPoint(const WebCore::IntPoint& point, WebCore::TextGranularity granularity, bool isInteractingWithFocusedElement)
        : m_arguments(point, granularity, isInteractingWithFocusedElement)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class SelectPositionAtBoundaryWithDirection {
public:
    using Arguments = std::tuple<const WebCore::IntPoint&, WebCore::TextGranularity, WebCore::SelectionDirection, bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SelectPositionAtBoundaryWithDirection; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_SelectPositionAtBoundaryWithDirectionReply; }
    using AsyncReply = SelectPositionAtBoundaryWithDirectionAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    SelectPositionAtBoundaryWithDirection(const WebCore::IntPoint& point, WebCore::TextGranularity granularity, WebCore::SelectionDirection direction, bool isInteractingWithFocusedElement)
        : m_arguments(point, granularity, direction, isInteractingWithFocusedElement)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class MoveSelectionAtBoundaryWithDirection {
public:
    using Arguments = std::tuple<WebCore::TextGranularity, WebCore::SelectionDirection>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_MoveSelectionAtBoundaryWithDirection; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_MoveSelectionAtBoundaryWithDirectionReply; }
    using AsyncReply = MoveSelectionAtBoundaryWithDirectionAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    MoveSelectionAtBoundaryWithDirection(WebCore::TextGranularity granularity, WebCore::SelectionDirection direction)
        : m_arguments(granularity, direction)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class SelectPositionAtPoint {
public:
    using Arguments = std::tuple<const WebCore::IntPoint&, bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SelectPositionAtPoint; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_SelectPositionAtPointReply; }
    using AsyncReply = SelectPositionAtPointAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    SelectPositionAtPoint(const WebCore::IntPoint& point, bool isInteractingWithFocusedElement)
        : m_arguments(point, isInteractingWithFocusedElement)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class BeginSelectionInDirection {
public:
    using Arguments = std::tuple<WebCore::SelectionDirection>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_BeginSelectionInDirection; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_BeginSelectionInDirectionReply; }
    using AsyncReply = BeginSelectionInDirectionAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool endIsMoving);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit BeginSelectionInDirection(WebCore::SelectionDirection direction)
        : m_arguments(direction)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class UpdateSelectionWithExtentPoint {
public:
    using Arguments = std::tuple<const WebCore::IntPoint&, bool, WebKit::RespectSelectionAnchor>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_UpdateSelectionWithExtentPoint; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_UpdateSelectionWithExtentPointReply; }
    using AsyncReply = UpdateSelectionWithExtentPointAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool endIsMoving);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    UpdateSelectionWithExtentPoint(const WebCore::IntPoint& point, bool isInteractingWithFocusedElement, WebKit::RespectSelectionAnchor respectSelectionAnchor)
        : m_arguments(point, isInteractingWithFocusedElement, respectSelectionAnchor)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class UpdateSelectionWithExtentPointAndBoundary {
public:
    using Arguments = std::tuple<const WebCore::IntPoint&, WebCore::TextGranularity, bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_UpdateSelectionWithExtentPointAndBoundary; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_UpdateSelectionWithExtentPointAndBoundaryReply; }
    using AsyncReply = UpdateSelectionWithExtentPointAndBoundaryAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool endIsMoving);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    UpdateSelectionWithExtentPointAndBoundary(const WebCore::IntPoint& point, WebCore::TextGranularity granularity, bool isInteractingWithFocusedElement)
        : m_arguments(point, granularity, isInteractingWithFocusedElement)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class RequestDictationContext {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RequestDictationContext; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(String&&, String&&, String&&)>&&);
    static void cancelReply(CompletionHandler<void(String&&, String&&, String&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_RequestDictationContextReply; }
    using AsyncReply = RequestDictationContextAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& selectedText, const String& textBefore, const String& textAfter);
    using Reply = std::tuple<String&, String&, String&>;
    using ReplyArguments = std::tuple<String, String, String>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class ReplaceDictatedText {
public:
    using Arguments = std::tuple<const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ReplaceDictatedText; }
    static const bool isSync = false;

    ReplaceDictatedText(const String& oldText, const String& newText)
        : m_arguments(oldText, newText)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class ReplaceSelectedText {
public:
    using Arguments = std::tuple<const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ReplaceSelectedText; }
    static const bool isSync = false;

    ReplaceSelectedText(const String& oldText, const String& newText)
        : m_arguments(oldText, newText)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class RequestAutocorrectionData {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RequestAutocorrectionData; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebKit::WebAutocorrectionData&&)>&&);
    static void cancelReply(CompletionHandler<void(WebKit::WebAutocorrectionData&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_RequestAutocorrectionDataReply; }
    using AsyncReply = RequestAutocorrectionDataAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::WebAutocorrectionData& data);
    using Reply = std::tuple<WebKit::WebAutocorrectionData&>;
    using ReplyArguments = std::tuple<WebKit::WebAutocorrectionData>;
    explicit RequestAutocorrectionData(const String& textForAutocorrection)
        : m_arguments(textForAutocorrection)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class ApplyAutocorrection {
public:
    using Arguments = std::tuple<const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ApplyAutocorrection; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(String&&)>&&);
    static void cancelReply(CompletionHandler<void(String&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_ApplyAutocorrectionReply; }
    using AsyncReply = ApplyAutocorrectionAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& string);
    using Reply = std::tuple<String&>;
    using ReplyArguments = std::tuple<String>;
    ApplyAutocorrection(const String& correction, const String& originalText)
        : m_arguments(correction, originalText)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class SyncApplyAutocorrection {
public:
    using Arguments = std::tuple<const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SyncApplyAutocorrection; }
    static const bool isSync = true;

    using DelayedReply = SyncApplyAutocorrectionDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool autocorrectionApplied);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    SyncApplyAutocorrection(const String& correction, const String& originalText)
        : m_arguments(correction, originalText)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class RequestAutocorrectionContext {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RequestAutocorrectionContext; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class RequestEvasionRectsAboveSelection {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RequestEvasionRectsAboveSelection; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Vector<WebCore::FloatRect>&&)>&&);
    static void cancelReply(CompletionHandler<void(Vector<WebCore::FloatRect>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_RequestEvasionRectsAboveSelectionReply; }
    using AsyncReply = RequestEvasionRectsAboveSelectionAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<WebCore::FloatRect>& rects);
    using Reply = std::tuple<Vector<WebCore::FloatRect>&>;
    using ReplyArguments = std::tuple<Vector<WebCore::FloatRect>>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class GetPositionInformation {
public:
    using Arguments = std::tuple<const WebKit::InteractionInformationRequest&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetPositionInformation; }
    static const bool isSync = true;

    using DelayedReply = GetPositionInformationDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::InteractionInformationAtPosition& information);
    using Reply = std::tuple<WebKit::InteractionInformationAtPosition&>;
    using ReplyArguments = std::tuple<WebKit::InteractionInformationAtPosition>;
    explicit GetPositionInformation(const WebKit::InteractionInformationRequest& request)
        : m_arguments(request)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class RequestPositionInformation {
public:
    using Arguments = std::tuple<const WebKit::InteractionInformationRequest&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RequestPositionInformation; }
    static const bool isSync = false;

    explicit RequestPositionInformation(const WebKit::InteractionInformationRequest& request)
        : m_arguments(request)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class StartInteractionWithElementContextOrPosition {
public:
    using Arguments = std::tuple<const Optional<WebCore::ElementContext>&, const WebCore::IntPoint&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_StartInteractionWithElementContextOrPosition; }
    static const bool isSync = false;

    StartInteractionWithElementContextOrPosition(const Optional<WebCore::ElementContext>& elementContext, const WebCore::IntPoint& point)
        : m_arguments(elementContext, point)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class StopInteraction {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_StopInteraction; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class PerformActionOnElement {
public:
    using Arguments = std::tuple<uint32_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_PerformActionOnElement; }
    static const bool isSync = false;

    explicit PerformActionOnElement(uint32_t action)
        : m_arguments(action)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class FocusNextFocusedElement {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_FocusNextFocusedElement; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_FocusNextFocusedElementReply; }
    using AsyncReply = FocusNextFocusedElementAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    explicit FocusNextFocusedElement(bool isForward)
        : m_arguments(isForward)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class SetFocusedElementValue {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetFocusedElementValue; }
    static const bool isSync = false;

    explicit SetFocusedElementValue(const String& value)
        : m_arguments(value)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class AutofillLoginCredentials {
public:
    using Arguments = std::tuple<const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_AutofillLoginCredentials; }
    static const bool isSync = false;

    AutofillLoginCredentials(const String& username, const String& password)
        : m_arguments(username, password)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class SetFocusedElementValueAsNumber {
public:
    using Arguments = std::tuple<double>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetFocusedElementValueAsNumber; }
    static const bool isSync = false;

    explicit SetFocusedElementValueAsNumber(double value)
        : m_arguments(value)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class SetFocusedElementSelectedIndex {
public:
    using Arguments = std::tuple<uint32_t, bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetFocusedElementSelectedIndex; }
    static const bool isSync = false;

    SetFocusedElementSelectedIndex(uint32_t index, bool allowMultipleSelection)
        : m_arguments(index, allowMultipleSelection)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class ApplicationWillResignActive {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ApplicationWillResignActive; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class ApplicationDidEnterBackground {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ApplicationDidEnterBackground; }
    static const bool isSync = false;

    explicit ApplicationDidEnterBackground(bool isSuspendedUnderLock)
        : m_arguments(isSuspendedUnderLock)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class ApplicationDidFinishSnapshottingAfterEnteringBackground {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ApplicationDidFinishSnapshottingAfterEnteringBackground; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class ApplicationWillEnterForeground {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ApplicationWillEnterForeground; }
    static const bool isSync = false;

    explicit ApplicationWillEnterForeground(bool isSuspendedUnderLock)
        : m_arguments(isSuspendedUnderLock)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class ApplicationDidBecomeActive {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ApplicationDidBecomeActive; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class ApplicationDidEnterBackgroundForMedia {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ApplicationDidEnterBackgroundForMedia; }
    static const bool isSync = false;

    explicit ApplicationDidEnterBackgroundForMedia(bool isSuspendedUnderLock)
        : m_arguments(isSuspendedUnderLock)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class ApplicationWillEnterForegroundForMedia {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ApplicationWillEnterForegroundForMedia; }
    static const bool isSync = false;

    explicit ApplicationWillEnterForegroundForMedia(bool isSuspendedUnderLock)
        : m_arguments(isSuspendedUnderLock)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class ContentSizeCategoryDidChange {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ContentSizeCategoryDidChange; }
    static const bool isSync = false;

    explicit ContentSizeCategoryDidChange(const String& contentSizeCategory)
        : m_arguments(contentSizeCategory)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class GetSelectionContext {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetSelectionContext; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(String&&, String&&, String&&)>&&);
    static void cancelReply(CompletionHandler<void(String&&, String&&, String&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetSelectionContextReply; }
    using AsyncReply = GetSelectionContextAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& selectedText, const String& textBefore, const String& textAfter);
    using Reply = std::tuple<String&, String&, String&>;
    using ReplyArguments = std::tuple<String, String, String>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class SetAllowsMediaDocumentInlinePlayback {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetAllowsMediaDocumentInlinePlayback; }
    static const bool isSync = false;

    explicit SetAllowsMediaDocumentInlinePlayback(bool allows)
        : m_arguments(allows)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class HandleTwoFingerTapAtPoint {
public:
    using Arguments = std::tuple<const WebCore::IntPoint&, const OptionSet<WebKit::WebEvent::Modifier>&, uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_HandleTwoFingerTapAtPoint; }
    static const bool isSync = false;

    HandleTwoFingerTapAtPoint(const WebCore::IntPoint& point, const OptionSet<WebKit::WebEvent::Modifier>& modifiers, uint64_t requestID)
        : m_arguments(point, modifiers, requestID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class SetForceAlwaysUserScalable {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetForceAlwaysUserScalable; }
    static const bool isSync = false;

    explicit SetForceAlwaysUserScalable(bool userScalable)
        : m_arguments(userScalable)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class GetRectsForGranularityWithSelectionOffset {
public:
    using Arguments = std::tuple<WebCore::TextGranularity, int32_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetRectsForGranularityWithSelectionOffset; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Vector<WebCore::SelectionRect>&&)>&&);
    static void cancelReply(CompletionHandler<void(Vector<WebCore::SelectionRect>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetRectsForGranularityWithSelectionOffsetReply; }
    using AsyncReply = GetRectsForGranularityWithSelectionOffsetAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<WebCore::SelectionRect>& rect);
    using Reply = std::tuple<Vector<WebCore::SelectionRect>&>;
    using ReplyArguments = std::tuple<Vector<WebCore::SelectionRect>>;
    GetRectsForGranularityWithSelectionOffset(WebCore::TextGranularity granularity, int32_t offset)
        : m_arguments(granularity, offset)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class GetRectsAtSelectionOffsetWithText {
public:
    using Arguments = std::tuple<int32_t, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetRectsAtSelectionOffsetWithText; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Vector<WebCore::SelectionRect>&&)>&&);
    static void cancelReply(CompletionHandler<void(Vector<WebCore::SelectionRect>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetRectsAtSelectionOffsetWithTextReply; }
    using AsyncReply = GetRectsAtSelectionOffsetWithTextAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<WebCore::SelectionRect>& rect);
    using Reply = std::tuple<Vector<WebCore::SelectionRect>&>;
    using ReplyArguments = std::tuple<Vector<WebCore::SelectionRect>>;
    GetRectsAtSelectionOffsetWithText(int32_t offset, const String& text)
        : m_arguments(offset, text)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class StoreSelectionForAccessibility {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_StoreSelectionForAccessibility; }
    static const bool isSync = false;

    explicit StoreSelectionForAccessibility(bool shouldStore)
        : m_arguments(shouldStore)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class StartAutoscrollAtPosition {
public:
    using Arguments = std::tuple<const WebCore::FloatPoint&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_StartAutoscrollAtPosition; }
    static const bool isSync = false;

    explicit StartAutoscrollAtPosition(const WebCore::FloatPoint& positionInWindow)
        : m_arguments(positionInWindow)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class CancelAutoscroll {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_CancelAutoscroll; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class RequestFocusedElementInformation {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RequestFocusedElementInformation; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebKit::FocusedElementInformation&&)>&&);
    static void cancelReply(CompletionHandler<void(WebKit::FocusedElementInformation&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_RequestFocusedElementInformationReply; }
    using AsyncReply = RequestFocusedElementInformationAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::FocusedElementInformation& info);
    using Reply = std::tuple<WebKit::FocusedElementInformation&>;
    using ReplyArguments = std::tuple<WebKit::FocusedElementInformation>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class HardwareKeyboardAvailabilityChanged {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_HardwareKeyboardAvailabilityChanged; }
    static const bool isSync = false;

    explicit HardwareKeyboardAvailabilityChanged(bool keyboardIsAttached)
        : m_arguments(keyboardIsAttached)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class SetIsShowingInputViewForFocusedElement {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetIsShowingInputViewForFocusedElement; }
    static const bool isSync = false;

    explicit SetIsShowingInputViewForFocusedElement(bool showingInputView)
        : m_arguments(showingInputView)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class UpdateSelectionWithDelta {
public:
    using Arguments = std::tuple<int64_t, int64_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_UpdateSelectionWithDelta; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_UpdateSelectionWithDeltaReply; }
    using AsyncReply = UpdateSelectionWithDeltaAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    UpdateSelectionWithDelta(int64_t locationDelta, int64_t lengthDelta)
        : m_arguments(locationDelta, lengthDelta)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class RequestDocumentEditingContext {
public:
    using Arguments = std::tuple<const WebKit::DocumentEditingContextRequest&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RequestDocumentEditingContext; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebKit::DocumentEditingContext&&)>&&);
    static void cancelReply(CompletionHandler<void(WebKit::DocumentEditingContext&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_RequestDocumentEditingContextReply; }
    using AsyncReply = RequestDocumentEditingContextAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::DocumentEditingContext& response);
    using Reply = std::tuple<WebKit::DocumentEditingContext&>;
    using ReplyArguments = std::tuple<WebKit::DocumentEditingContext>;
    explicit RequestDocumentEditingContext(const WebKit::DocumentEditingContextRequest& request)
        : m_arguments(request)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class GenerateSyntheticEditingCommand {
public:
    using Arguments = std::tuple<WebKit::SyntheticEditingCommandType>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GenerateSyntheticEditingCommand; }
    static const bool isSync = false;

    explicit GenerateSyntheticEditingCommand(WebKit::SyntheticEditingCommandType command)
        : m_arguments(command)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class SetShouldRevealCurrentSelectionAfterInsertion {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetShouldRevealCurrentSelectionAfterInsertion; }
    static const bool isSync = false;

    explicit SetShouldRevealCurrentSelectionAfterInsertion(bool shouldRevealCurrentSelectionAfterInsertion)
        : m_arguments(shouldRevealCurrentSelectionAfterInsertion)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class InsertTextPlaceholder {
public:
    using Arguments = std::tuple<const WebCore::IntSize&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_InsertTextPlaceholder; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Optional<WebCore::ElementContext>&&)>&&);
    static void cancelReply(CompletionHandler<void(Optional<WebCore::ElementContext>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_InsertTextPlaceholderReply; }
    using AsyncReply = InsertTextPlaceholderAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<WebCore::ElementContext>& placeholder);
    using Reply = std::tuple<Optional<WebCore::ElementContext>&>;
    using ReplyArguments = std::tuple<Optional<WebCore::ElementContext>>;
    explicit InsertTextPlaceholder(const WebCore::IntSize& size)
        : m_arguments(size)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class RemoveTextPlaceholder {
public:
    using Arguments = std::tuple<const WebCore::ElementContext&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RemoveTextPlaceholder; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_RemoveTextPlaceholderReply; }
    using AsyncReply = RemoveTextPlaceholderAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    explicit RemoveTextPlaceholder(const WebCore::ElementContext& placeholder)
        : m_arguments(placeholder)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class TextInputContextsInRect {
public:
    using Arguments = std::tuple<const WebCore::FloatRect&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_TextInputContextsInRect; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Vector<struct WebCore::ElementContext>&&)>&&);
    static void cancelReply(CompletionHandler<void(Vector<struct WebCore::ElementContext>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_TextInputContextsInRectReply; }
    using AsyncReply = TextInputContextsInRectAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<struct WebCore::ElementContext>& contexts);
    using Reply = std::tuple<Vector<struct WebCore::ElementContext>&>;
    using ReplyArguments = std::tuple<Vector<struct WebCore::ElementContext>>;
    explicit TextInputContextsInRect(const WebCore::FloatRect& rect)
        : m_arguments(rect)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class FocusTextInputContextAndPlaceCaret {
public:
    using Arguments = std::tuple<const WebCore::ElementContext&, const WebCore::IntPoint&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_FocusTextInputContextAndPlaceCaret; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_FocusTextInputContextAndPlaceCaretReply; }
    using AsyncReply = FocusTextInputContextAndPlaceCaretAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool success);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    FocusTextInputContextAndPlaceCaret(const WebCore::ElementContext& context, const WebCore::IntPoint& point)
        : m_arguments(context, point)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class ClearServiceWorkerEntitlementOverride {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ClearServiceWorkerEntitlementOverride; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_ClearServiceWorkerEntitlementOverrideReply; }
    using AsyncReply = ClearServiceWorkerEntitlementOverrideAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class SetControlledByAutomation {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetControlledByAutomation; }
    static const bool isSync = false;

    explicit SetControlledByAutomation(bool controlled)
        : m_arguments(controlled)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ConnectInspector {
public:
    using Arguments = std::tuple<const String&, const Inspector::FrontendChannel::ConnectionType&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ConnectInspector; }
    static const bool isSync = false;

    ConnectInspector(const String& targetId, const Inspector::FrontendChannel::ConnectionType& connectionType)
        : m_arguments(targetId, connectionType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DisconnectInspector {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DisconnectInspector; }
    static const bool isSync = false;

    explicit DisconnectInspector(const String& targetId)
        : m_arguments(targetId)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SendMessageToTargetBackend {
public:
    using Arguments = std::tuple<const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SendMessageToTargetBackend; }
    static const bool isSync = false;

    SendMessageToTargetBackend(const String& targetId, const String& message)
        : m_arguments(targetId, message)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if ENABLE(REMOTE_INSPECTOR)
class SetIndicating {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetIndicating; }
    static const bool isSync = false;

    explicit SetIndicating(bool indicating)
        : m_arguments(indicating)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(IOS_TOUCH_EVENTS)
class ResetPotentialTapSecurityOrigin {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ResetPotentialTapSecurityOrigin; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(IOS_TOUCH_EVENTS)
class TouchEventSync {
public:
    using Arguments = std::tuple<const WebKit::WebTouchEvent&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_TouchEventSync; }
    static const bool isSync = true;

    using DelayedReply = TouchEventSyncDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool handled);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit TouchEventSync(const WebKit::WebTouchEvent& event)
        : m_arguments(event)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if !ENABLE(IOS_TOUCH_EVENTS) && ENABLE(TOUCH_EVENTS)
class TouchEvent {
public:
    using Arguments = std::tuple<const WebKit::WebTouchEvent&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_TouchEvent; }
    static const bool isSync = false;

    explicit TouchEvent(const WebKit::WebTouchEvent& event)
        : m_arguments(event)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class CancelPointer {
public:
    using Arguments = std::tuple<const WebCore::PointerID&, const WebCore::IntPoint&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_CancelPointer; }
    static const bool isSync = false;

    CancelPointer(const WebCore::PointerID& pointerId, const WebCore::IntPoint& documentPoint)
        : m_arguments(pointerId, documentPoint)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class TouchWithIdentifierWasRemoved {
public:
    using Arguments = std::tuple<const WebCore::PointerID&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_TouchWithIdentifierWasRemoved; }
    static const bool isSync = false;

    explicit TouchWithIdentifierWasRemoved(const WebCore::PointerID& pointerId)
        : m_arguments(pointerId)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if ENABLE(INPUT_TYPE_COLOR)
class DidEndColorPicker {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidEndColorPicker; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(INPUT_TYPE_COLOR)
class DidChooseColor {
public:
    using Arguments = std::tuple<const WebCore::Color&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidChooseColor; }
    static const bool isSync = false;

    explicit DidChooseColor(const WebCore::Color& color)
        : m_arguments(color)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(DATALIST_ELEMENT)
class DidSelectDataListOption {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidSelectDataListOption; }
    static const bool isSync = false;

    explicit DidSelectDataListOption(const String& selectedOption)
        : m_arguments(selectedOption)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(DATALIST_ELEMENT)
class DidCloseSuggestions {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidCloseSuggestions; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(DATE_AND_TIME_INPUT_TYPES)
class DidChooseDate {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidChooseDate; }
    static const bool isSync = false;

    explicit DidChooseDate(const String& date)
        : m_arguments(date)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(DATE_AND_TIME_INPUT_TYPES)
class DidEndDateTimePicker {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidEndDateTimePicker; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(CONTEXT_MENUS)
class ContextMenuHidden {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ContextMenuHidden; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(CONTEXT_MENUS)
class ContextMenuForKeyEvent {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ContextMenuForKeyEvent; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class ScrollBy {
public:
    using Arguments = std::tuple<uint32_t, uint32_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ScrollBy; }
    static const bool isSync = false;

    ScrollBy(uint32_t scrollDirection, uint32_t scrollGranularity)
        : m_arguments(scrollDirection, scrollGranularity)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CenterSelectionInVisibleArea {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_CenterSelectionInVisibleArea; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GoToBackForwardItem {
public:
    using Arguments = std::tuple<uint64_t, const WebCore::BackForwardItemIdentifier&, WebCore::FrameLoadType, WebCore::ShouldTreatAsContinuingLoad, const Optional<WebKit::WebsitePoliciesData>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GoToBackForwardItem; }
    static const bool isSync = false;

    GoToBackForwardItem(uint64_t navigationID, const WebCore::BackForwardItemIdentifier& backForwardItemID, WebCore::FrameLoadType backForwardType, WebCore::ShouldTreatAsContinuingLoad shouldTreatAsContinuingLoad, const Optional<WebKit::WebsitePoliciesData>& websitePolicies)
        : m_arguments(navigationID, backForwardItemID, backForwardType, shouldTreatAsContinuingLoad, websitePolicies)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class TryRestoreScrollPosition {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_TryRestoreScrollPosition; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class LoadURLInFrame {
public:
    using Arguments = std::tuple<const URL&, const String&, const WebCore::FrameIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_LoadURLInFrame; }
    static const bool isSync = false;

    LoadURLInFrame(const URL& url, const String& referrer, const WebCore::FrameIdentifier& frameID)
        : m_arguments(url, referrer, frameID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class LoadDataInFrame {
public:
    using Arguments = std::tuple<const IPC::DataReference&, const String&, const String&, const URL&, const WebCore::FrameIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_LoadDataInFrame; }
    static const bool isSync = false;

    LoadDataInFrame(const IPC::DataReference& data, const String& MIMEType, const String& encodingName, const URL& baseURL, const WebCore::FrameIdentifier& frameID)
        : m_arguments(data, MIMEType, encodingName, baseURL, frameID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class LoadRequest {
public:
    using Arguments = std::tuple<const WebKit::LoadParameters&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_LoadRequest; }
    static const bool isSync = false;

    explicit LoadRequest(const WebKit::LoadParameters& loadParameters)
        : m_arguments(loadParameters)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class LoadRequestWaitingForProcessLaunch {
public:
    using Arguments = std::tuple<const WebKit::LoadParameters&, const URL&, const WebKit::WebPageProxyIdentifier&, bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_LoadRequestWaitingForProcessLaunch; }
    static const bool isSync = false;

    LoadRequestWaitingForProcessLaunch(const WebKit::LoadParameters& loadParameters, const URL& resourceDirectoryURL, const WebKit::WebPageProxyIdentifier& pageID, bool checkAssumedReadAccessToResourceURL)
        : m_arguments(loadParameters, resourceDirectoryURL, pageID, checkAssumedReadAccessToResourceURL)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class LoadData {
public:
    using Arguments = std::tuple<const WebKit::LoadParameters&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_LoadData; }
    static const bool isSync = false;

    explicit LoadData(const WebKit::LoadParameters& loadParameters)
        : m_arguments(loadParameters)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class LoadSimulatedRequestAndResponse {
public:
    using Arguments = std::tuple<const WebKit::LoadParameters&, const WebCore::ResourceResponse&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_LoadSimulatedRequestAndResponse; }
    static const bool isSync = false;

    LoadSimulatedRequestAndResponse(const WebKit::LoadParameters& loadParameters, const WebCore::ResourceResponse& simulatedResponse)
        : m_arguments(loadParameters, simulatedResponse)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class LoadAlternateHTML {
public:
    using Arguments = std::tuple<const WebKit::LoadParameters&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_LoadAlternateHTML; }
    static const bool isSync = false;

    explicit LoadAlternateHTML(const WebKit::LoadParameters& loadParameters)
        : m_arguments(loadParameters)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class NavigateToPDFLinkWithSimulatedClick {
public:
    using Arguments = std::tuple<const String&, const WebCore::IntPoint&, const WebCore::IntPoint&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_NavigateToPDFLinkWithSimulatedClick; }
    static const bool isSync = false;

    NavigateToPDFLinkWithSimulatedClick(const String& url, const WebCore::IntPoint& documentPoint, const WebCore::IntPoint& screenPoint)
        : m_arguments(url, documentPoint, screenPoint)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetPDFFirstPageSize {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetPDFFirstPageSize; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebCore::FloatSize&&)>&&);
    static void cancelReply(CompletionHandler<void(WebCore::FloatSize&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetPDFFirstPageSizeReply; }
    using AsyncReply = GetPDFFirstPageSizeAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebCore::FloatSize& size);
    using Reply = std::tuple<WebCore::FloatSize&>;
    using ReplyArguments = std::tuple<WebCore::FloatSize>;
    explicit GetPDFFirstPageSize(const WebCore::FrameIdentifier& frameID)
        : m_arguments(frameID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class Reload {
public:
    using Arguments = std::tuple<uint64_t, uint32_t, const WebKit::SandboxExtension::Handle&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_Reload; }
    static const bool isSync = false;

    Reload(uint64_t navigationID, uint32_t reloadOptions, const WebKit::SandboxExtension::Handle& sandboxExtensionHandle)
        : m_arguments(navigationID, reloadOptions, sandboxExtensionHandle)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class StopLoading {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_StopLoading; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class StopLoadingFrame {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_StopLoadingFrame; }
    static const bool isSync = false;

    explicit StopLoadingFrame(const WebCore::FrameIdentifier& frameID)
        : m_arguments(frameID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RestoreSession {
public:
    using Arguments = std::tuple<const Vector<WebKit::BackForwardListItemState>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RestoreSession; }
    static const bool isSync = false;

    explicit RestoreSession(const Vector<WebKit::BackForwardListItemState>& itemStates)
        : m_arguments(itemStates)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class UpdateBackForwardListForReattach {
public:
    using Arguments = std::tuple<const Vector<WebKit::BackForwardListItemState>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_UpdateBackForwardListForReattach; }
    static const bool isSync = false;

    explicit UpdateBackForwardListForReattach(const Vector<WebKit::BackForwardListItemState>& itemStates)
        : m_arguments(itemStates)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetCurrentHistoryItemForReattach {
public:
    using Arguments = std::tuple<const WebKit::BackForwardListItemState&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetCurrentHistoryItemForReattach; }
    static const bool isSync = false;

    explicit SetCurrentHistoryItemForReattach(const WebKit::BackForwardListItemState& itemState)
        : m_arguments(itemState)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidRemoveBackForwardItem {
public:
    using Arguments = std::tuple<const WebCore::BackForwardItemIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidRemoveBackForwardItem; }
    static const bool isSync = false;

    explicit DidRemoveBackForwardItem(const WebCore::BackForwardItemIdentifier& backForwardItemID)
        : m_arguments(backForwardItemID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class UpdateWebsitePolicies {
public:
    using Arguments = std::tuple<const WebKit::WebsitePoliciesData&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_UpdateWebsitePolicies; }
    static const bool isSync = false;

    explicit UpdateWebsitePolicies(const WebKit::WebsitePoliciesData& websitePolicies)
        : m_arguments(websitePolicies)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class NotifyUserScripts {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_NotifyUserScripts; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidReceivePolicyDecision {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&, uint64_t, const WebKit::PolicyDecision&, const WebKit::SandboxExtension::HandleArray&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidReceivePolicyDecision; }
    static const bool isSync = false;

    DidReceivePolicyDecision(const WebCore::FrameIdentifier& frameID, uint64_t listenerID, const WebKit::PolicyDecision& policyDecision, const WebKit::SandboxExtension::HandleArray& networkExtensionsSandboxExtensions)
        : m_arguments(frameID, listenerID, policyDecision, networkExtensionsSandboxExtensions)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ContinueWillSubmitForm {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&, uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ContinueWillSubmitForm; }
    static const bool isSync = false;

    ContinueWillSubmitForm(const WebCore::FrameIdentifier& frameID, uint64_t listenerID)
        : m_arguments(frameID, listenerID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ClearSelection {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ClearSelection; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RestoreSelectionInFocusedEditableElement {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RestoreSelectionInFocusedEditableElement; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetContentsAsString {
public:
    using Arguments = std::tuple<WebKit::ContentAsStringIncludesChildFrames>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetContentsAsString; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(String&&)>&&);
    static void cancelReply(CompletionHandler<void(String&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetContentsAsStringReply; }
    using AsyncReply = GetContentsAsStringAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& string);
    using Reply = std::tuple<String&>;
    using ReplyArguments = std::tuple<String>;
    explicit GetContentsAsString(WebKit::ContentAsStringIncludesChildFrames inChildFrames)
        : m_arguments(inChildFrames)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetAllFrames {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetAllFrames; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebKit::FrameTreeNodeData&&)>&&);
    static void cancelReply(CompletionHandler<void(WebKit::FrameTreeNodeData&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetAllFramesReply; }
    using AsyncReply = GetAllFramesAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::FrameTreeNodeData& mainFrame);
    using Reply = std::tuple<WebKit::FrameTreeNodeData&>;
    using ReplyArguments = std::tuple<WebKit::FrameTreeNodeData>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(COCOA)
class GetContentsAsAttributedString {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetContentsAsAttributedString; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebCore::AttributedString&&)>&&);
    static void cancelReply(CompletionHandler<void(WebCore::AttributedString&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetContentsAsAttributedStringReply; }
    using AsyncReply = GetContentsAsAttributedStringAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebCore::AttributedString& result);
    using Reply = std::tuple<WebCore::AttributedString&>;
    using ReplyArguments = std::tuple<WebCore::AttributedString>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(MHTML)
class GetContentsAsMHTMLData {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetContentsAsMHTMLData; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(IPC::DataReference&&)>&&);
    static void cancelReply(CompletionHandler<void(IPC::DataReference&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetContentsAsMHTMLDataReply; }
    using AsyncReply = GetContentsAsMHTMLDataAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const IPC::SharedBufferDataReference& data);
    using Reply = std::tuple<IPC::SharedBufferDataReference&>;
    using ReplyArguments = std::tuple<IPC::DataReference>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class GetMainResourceDataOfFrame {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetMainResourceDataOfFrame; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Optional<IPC::DataReference>&&)>&&);
    static void cancelReply(CompletionHandler<void(Optional<IPC::DataReference>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetMainResourceDataOfFrameReply; }
    using AsyncReply = GetMainResourceDataOfFrameAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<IPC::SharedBufferDataReference>& dataReference);
    using Reply = std::tuple<Optional<IPC::SharedBufferDataReference>&>;
    using ReplyArguments = std::tuple<Optional<IPC::DataReference>>;
    explicit GetMainResourceDataOfFrame(const WebCore::FrameIdentifier& frameID)
        : m_arguments(frameID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetResourceDataFromFrame {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetResourceDataFromFrame; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Optional<IPC::DataReference>&&)>&&);
    static void cancelReply(CompletionHandler<void(Optional<IPC::DataReference>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetResourceDataFromFrameReply; }
    using AsyncReply = GetResourceDataFromFrameAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<IPC::SharedBufferDataReference>& dataReference);
    using Reply = std::tuple<Optional<IPC::SharedBufferDataReference>&>;
    using ReplyArguments = std::tuple<Optional<IPC::DataReference>>;
    GetResourceDataFromFrame(const WebCore::FrameIdentifier& frameID, const String& resourceURL)
        : m_arguments(frameID, resourceURL)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetRenderTreeExternalRepresentation {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetRenderTreeExternalRepresentation; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(String&&)>&&);
    static void cancelReply(CompletionHandler<void(String&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetRenderTreeExternalRepresentationReply; }
    using AsyncReply = GetRenderTreeExternalRepresentationAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& string);
    using Reply = std::tuple<String&>;
    using ReplyArguments = std::tuple<String>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetSelectionOrContentsAsString {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetSelectionOrContentsAsString; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(String&&)>&&);
    static void cancelReply(CompletionHandler<void(String&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetSelectionOrContentsAsStringReply; }
    using AsyncReply = GetSelectionOrContentsAsStringAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& string);
    using Reply = std::tuple<String&>;
    using ReplyArguments = std::tuple<String>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetSelectionAsWebArchiveData {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetSelectionAsWebArchiveData; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Optional<IPC::DataReference>&&)>&&);
    static void cancelReply(CompletionHandler<void(Optional<IPC::DataReference>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetSelectionAsWebArchiveDataReply; }
    using AsyncReply = GetSelectionAsWebArchiveDataAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<IPC::DataReference>& data);
    using Reply = std::tuple<Optional<IPC::DataReference>&>;
    using ReplyArguments = std::tuple<Optional<IPC::DataReference>>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetSourceForFrame {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetSourceForFrame; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(String&&)>&&);
    static void cancelReply(CompletionHandler<void(String&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetSourceForFrameReply; }
    using AsyncReply = GetSourceForFrameAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& string);
    using Reply = std::tuple<String&>;
    using ReplyArguments = std::tuple<String>;
    explicit GetSourceForFrame(const WebCore::FrameIdentifier& frameID)
        : m_arguments(frameID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetWebArchiveOfFrame {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetWebArchiveOfFrame; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Optional<IPC::DataReference>&&)>&&);
    static void cancelReply(CompletionHandler<void(Optional<IPC::DataReference>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetWebArchiveOfFrameReply; }
    using AsyncReply = GetWebArchiveOfFrameAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<IPC::DataReference>& dataReference);
    using Reply = std::tuple<Optional<IPC::DataReference>&>;
    using ReplyArguments = std::tuple<Optional<IPC::DataReference>>;
    explicit GetWebArchiveOfFrame(const WebCore::FrameIdentifier& frameID)
        : m_arguments(frameID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RunJavaScriptInFrameInScriptWorld {
public:
    using Arguments = std::tuple<const WebCore::RunJavaScriptParameters&, const Optional<WebCore::FrameIdentifier>&, const std::pair<WebKit::ContentWorldIdentifier, String>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RunJavaScriptInFrameInScriptWorld; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(IPC::DataReference&&, Optional<WebCore::ExceptionDetails>&&)>&&);
    static void cancelReply(CompletionHandler<void(IPC::DataReference&&, Optional<WebCore::ExceptionDetails>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_RunJavaScriptInFrameInScriptWorldReply; }
    using AsyncReply = RunJavaScriptInFrameInScriptWorldAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const IPC::DataReference& resultData, const Optional<WebCore::ExceptionDetails>& details);
    using Reply = std::tuple<IPC::DataReference&, Optional<WebCore::ExceptionDetails>&>;
    using ReplyArguments = std::tuple<IPC::DataReference, Optional<WebCore::ExceptionDetails>>;
    RunJavaScriptInFrameInScriptWorld(const WebCore::RunJavaScriptParameters& parameters, const Optional<WebCore::FrameIdentifier>& frameID, const std::pair<WebKit::ContentWorldIdentifier, String>& world)
        : m_arguments(parameters, frameID, world)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ForceRepaint {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ForceRepaint; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_ForceRepaintReply; }
    using AsyncReply = ForceRepaintAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SelectAll {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SelectAll; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ScheduleFullEditorStateUpdate {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ScheduleFullEditorStateUpdate; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(COCOA)
class PerformDictionaryLookupOfCurrentSelection {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_PerformDictionaryLookupOfCurrentSelection; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class PerformDictionaryLookupAtLocation {
public:
    using Arguments = std::tuple<const WebCore::FloatPoint&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_PerformDictionaryLookupAtLocation; }
    static const bool isSync = false;

    explicit PerformDictionaryLookupAtLocation(const WebCore::FloatPoint& point)
        : m_arguments(point)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(DATA_DETECTION)
class DetectDataInAllFrames {
public:
    using Arguments = std::tuple<uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DetectDataInAllFrames; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebKit::DataDetectionResult&&)>&&);
    static void cancelReply(CompletionHandler<void(WebKit::DataDetectionResult&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_DetectDataInAllFramesReply; }
    using AsyncReply = DetectDataInAllFramesAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::DataDetectionResult& result);
    using Reply = std::tuple<WebKit::DataDetectionResult&>;
    using ReplyArguments = std::tuple<WebKit::DataDetectionResult>;
    explicit DetectDataInAllFrames(uint64_t types)
        : m_arguments(types)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(DATA_DETECTION)
class RemoveDataDetectedLinks {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RemoveDataDetectedLinks; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebKit::DataDetectionResult&&)>&&);
    static void cancelReply(CompletionHandler<void(WebKit::DataDetectionResult&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_RemoveDataDetectedLinksReply; }
    using AsyncReply = RemoveDataDetectedLinksAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::DataDetectionResult& result);
    using Reply = std::tuple<WebKit::DataDetectionResult&>;
    using ReplyArguments = std::tuple<WebKit::DataDetectionResult>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class ChangeFont {
public:
    using Arguments = std::tuple<const WebCore::FontChanges&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ChangeFont; }
    static const bool isSync = false;

    explicit ChangeFont(const WebCore::FontChanges& changes)
        : m_arguments(changes)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ChangeFontAttributes {
public:
    using Arguments = std::tuple<const WebCore::FontAttributeChanges&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ChangeFontAttributes; }
    static const bool isSync = false;

    explicit ChangeFontAttributes(const WebCore::FontAttributeChanges& changes)
        : m_arguments(changes)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class PreferencesDidChange {
public:
    using Arguments = std::tuple<const WebKit::WebPreferencesStore&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_PreferencesDidChange; }
    static const bool isSync = false;

    explicit PreferencesDidChange(const WebKit::WebPreferencesStore& store)
        : m_arguments(store)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetUserAgent {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetUserAgent; }
    static const bool isSync = false;

    explicit SetUserAgent(const String& userAgent)
        : m_arguments(userAgent)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetCustomTextEncodingName {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetCustomTextEncodingName; }
    static const bool isSync = false;

    explicit SetCustomTextEncodingName(const String& encodingName)
        : m_arguments(encodingName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SuspendActiveDOMObjectsAndAnimations {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SuspendActiveDOMObjectsAndAnimations; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ResumeActiveDOMObjectsAndAnimations {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ResumeActiveDOMObjectsAndAnimations; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class Close {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_Close; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class TryClose {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_TryClose; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_TryCloseReply; }
    using AsyncReply = TryCloseAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool shouldClose);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetEditable {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetEditable; }
    static const bool isSync = false;

    explicit SetEditable(bool editable)
        : m_arguments(editable)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ValidateCommand {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ValidateCommand; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&, int32_t&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&, int32_t&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_ValidateCommandReply; }
    using AsyncReply = ValidateCommandAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool isEnabled, int32_t state);
    using Reply = std::tuple<bool&, int32_t&>;
    using ReplyArguments = std::tuple<bool, int32_t>;
    explicit ValidateCommand(const String& name)
        : m_arguments(name)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ExecuteEditCommand {
public:
    using Arguments = std::tuple<const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ExecuteEditCommand; }
    static const bool isSync = false;

    ExecuteEditCommand(const String& name, const String& argument)
        : m_arguments(name, argument)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class IncreaseListLevel {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_IncreaseListLevel; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DecreaseListLevel {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DecreaseListLevel; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ChangeListType {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ChangeListType; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetBaseWritingDirection {
public:
    using Arguments = std::tuple<WebCore::WritingDirection>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetBaseWritingDirection; }
    static const bool isSync = false;

    explicit SetBaseWritingDirection(WebCore::WritingDirection direction)
        : m_arguments(direction)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetNeedsFontAttributes {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetNeedsFontAttributes; }
    static const bool isSync = false;

    explicit SetNeedsFontAttributes(bool needsFontAttributes)
        : m_arguments(needsFontAttributes)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RequestFontAttributesAtSelectionStart {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RequestFontAttributesAtSelectionStart; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebCore::FontAttributes&&)>&&);
    static void cancelReply(CompletionHandler<void(WebCore::FontAttributes&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_RequestFontAttributesAtSelectionStartReply; }
    using AsyncReply = RequestFontAttributesAtSelectionStartAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebCore::FontAttributes& attributes);
    using Reply = std::tuple<WebCore::FontAttributes&>;
    using ReplyArguments = std::tuple<WebCore::FontAttributes>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidRemoveEditCommand {
public:
    using Arguments = std::tuple<uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidRemoveEditCommand; }
    static const bool isSync = false;

    explicit DidRemoveEditCommand(uint64_t commandID)
        : m_arguments(commandID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ReapplyEditCommand {
public:
    using Arguments = std::tuple<uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ReapplyEditCommand; }
    static const bool isSync = false;

    explicit ReapplyEditCommand(uint64_t commandID)
        : m_arguments(commandID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class UnapplyEditCommand {
public:
    using Arguments = std::tuple<uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_UnapplyEditCommand; }
    static const bool isSync = false;

    explicit UnapplyEditCommand(uint64_t commandID)
        : m_arguments(commandID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetPageAndTextZoomFactors {
public:
    using Arguments = std::tuple<double, double>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetPageAndTextZoomFactors; }
    static const bool isSync = false;

    SetPageAndTextZoomFactors(double pageZoomFactor, double textZoomFactor)
        : m_arguments(pageZoomFactor, textZoomFactor)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetPageZoomFactor {
public:
    using Arguments = std::tuple<double>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetPageZoomFactor; }
    static const bool isSync = false;

    explicit SetPageZoomFactor(double zoomFactor)
        : m_arguments(zoomFactor)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetTextZoomFactor {
public:
    using Arguments = std::tuple<double>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetTextZoomFactor; }
    static const bool isSync = false;

    explicit SetTextZoomFactor(double zoomFactor)
        : m_arguments(zoomFactor)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class WindowScreenDidChange {
public:
    using Arguments = std::tuple<uint32_t, const Optional<unsigned>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_WindowScreenDidChange; }
    static const bool isSync = false;

    WindowScreenDidChange(uint32_t displayID, const Optional<unsigned>& nominalFramesPerSecond)
        : m_arguments(displayID, nominalFramesPerSecond)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class AccessibilitySettingsDidChange {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_AccessibilitySettingsDidChange; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ScalePage {
public:
    using Arguments = std::tuple<double, const WebCore::IntPoint&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ScalePage; }
    static const bool isSync = false;

    ScalePage(double scale, const WebCore::IntPoint& origin)
        : m_arguments(scale, origin)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ScalePageInViewCoordinates {
public:
    using Arguments = std::tuple<double, const WebCore::IntPoint&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ScalePageInViewCoordinates; }
    static const bool isSync = false;

    ScalePageInViewCoordinates(double scale, const WebCore::IntPoint& centerInViewCoordinates)
        : m_arguments(scale, centerInViewCoordinates)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ScaleView {
public:
    using Arguments = std::tuple<double>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ScaleView; }
    static const bool isSync = false;

    explicit ScaleView(double scale)
        : m_arguments(scale)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetUseFixedLayout {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetUseFixedLayout; }
    static const bool isSync = false;

    explicit SetUseFixedLayout(bool fixed)
        : m_arguments(fixed)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetFixedLayoutSize {
public:
    using Arguments = std::tuple<const WebCore::IntSize&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetFixedLayoutSize; }
    static const bool isSync = false;

    explicit SetFixedLayoutSize(const WebCore::IntSize& size)
        : m_arguments(size)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ListenForLayoutMilestones {
public:
    using Arguments = std::tuple<const OptionSet<WebCore::LayoutMilestone>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ListenForLayoutMilestones; }
    static const bool isSync = false;

    explicit ListenForLayoutMilestones(const OptionSet<WebCore::LayoutMilestone>& layoutMilestones)
        : m_arguments(layoutMilestones)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetSuppressScrollbarAnimations {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetSuppressScrollbarAnimations; }
    static const bool isSync = false;

    explicit SetSuppressScrollbarAnimations(bool suppressAnimations)
        : m_arguments(suppressAnimations)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetEnableVerticalRubberBanding {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetEnableVerticalRubberBanding; }
    static const bool isSync = false;

    explicit SetEnableVerticalRubberBanding(bool enableVerticalRubberBanding)
        : m_arguments(enableVerticalRubberBanding)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetEnableHorizontalRubberBanding {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetEnableHorizontalRubberBanding; }
    static const bool isSync = false;

    explicit SetEnableHorizontalRubberBanding(bool enableHorizontalRubberBanding)
        : m_arguments(enableHorizontalRubberBanding)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetBackgroundExtendsBeyondPage {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetBackgroundExtendsBeyondPage; }
    static const bool isSync = false;

    explicit SetBackgroundExtendsBeyondPage(bool backgroundExtendsBeyondPage)
        : m_arguments(backgroundExtendsBeyondPage)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetPaginationMode {
public:
    using Arguments = std::tuple<uint32_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetPaginationMode; }
    static const bool isSync = false;

    explicit SetPaginationMode(uint32_t mode)
        : m_arguments(mode)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetPaginationBehavesLikeColumns {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetPaginationBehavesLikeColumns; }
    static const bool isSync = false;

    explicit SetPaginationBehavesLikeColumns(bool behavesLikeColumns)
        : m_arguments(behavesLikeColumns)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetPageLength {
public:
    using Arguments = std::tuple<double>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetPageLength; }
    static const bool isSync = false;

    explicit SetPageLength(double pageLength)
        : m_arguments(pageLength)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetGapBetweenPages {
public:
    using Arguments = std::tuple<double>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetGapBetweenPages; }
    static const bool isSync = false;

    explicit SetGapBetweenPages(double gap)
        : m_arguments(gap)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetPaginationLineGridEnabled {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetPaginationLineGridEnabled; }
    static const bool isSync = false;

    explicit SetPaginationLineGridEnabled(bool lineGridEnabled)
        : m_arguments(lineGridEnabled)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class PostInjectedBundleMessage {
public:
    using Arguments = std::tuple<const String&, const WebKit::UserData&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_PostInjectedBundleMessage; }
    static const bool isSync = false;

    PostInjectedBundleMessage(const String& messageName, const WebKit::UserData& messageBody)
        : m_arguments(messageName, messageBody)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class FindString {
public:
    using Arguments = std::tuple<const String&, const OptionSet<WebKit::FindOptions>&, const unsigned&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_FindString; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_FindStringReply; }
    using AsyncReply = FindStringAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool found);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    FindString(const String& string, const OptionSet<WebKit::FindOptions>& findOptions, const unsigned& maxMatchCount)
        : m_arguments(string, findOptions, maxMatchCount)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class FindStringMatches {
public:
    using Arguments = std::tuple<const String&, const OptionSet<WebKit::FindOptions>&, const unsigned&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_FindStringMatches; }
    static const bool isSync = false;

    FindStringMatches(const String& string, const OptionSet<WebKit::FindOptions>& findOptions, const unsigned& maxMatchCount)
        : m_arguments(string, findOptions, maxMatchCount)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetImageForFindMatch {
public:
    using Arguments = std::tuple<uint32_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetImageForFindMatch; }
    static const bool isSync = false;

    explicit GetImageForFindMatch(uint32_t matchIndex)
        : m_arguments(matchIndex)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SelectFindMatch {
public:
    using Arguments = std::tuple<uint32_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SelectFindMatch; }
    static const bool isSync = false;

    explicit SelectFindMatch(uint32_t matchIndex)
        : m_arguments(matchIndex)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class IndicateFindMatch {
public:
    using Arguments = std::tuple<uint32_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_IndicateFindMatch; }
    static const bool isSync = false;

    explicit IndicateFindMatch(uint32_t matchIndex)
        : m_arguments(matchIndex)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class HideFindUI {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_HideFindUI; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CountStringMatches {
public:
    using Arguments = std::tuple<const String&, const OptionSet<WebKit::FindOptions>&, const unsigned&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_CountStringMatches; }
    static const bool isSync = false;

    CountStringMatches(const String& string, const OptionSet<WebKit::FindOptions>& findOptions, const unsigned& maxMatchCount)
        : m_arguments(string, findOptions, maxMatchCount)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ReplaceMatches {
public:
    using Arguments = std::tuple<const Vector<uint32_t>&, const String&, bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ReplaceMatches; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(uint64_t&&)>&&);
    static void cancelReply(CompletionHandler<void(uint64_t&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_ReplaceMatchesReply; }
    using AsyncReply = ReplaceMatchesAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, uint64_t numberOfReplacements);
    using Reply = std::tuple<uint64_t&>;
    using ReplyArguments = std::tuple<uint64_t>;
    ReplaceMatches(const Vector<uint32_t>& matchIndices, const String& replacementText, bool selectionOnly)
        : m_arguments(matchIndices, replacementText, selectionOnly)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class AddMIMETypeWithCustomContentProvider {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_AddMIMETypeWithCustomContentProvider; }
    static const bool isSync = false;

    explicit AddMIMETypeWithCustomContentProvider(const String& mimeType)
        : m_arguments(mimeType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(GTK) && ENABLE(DRAG_SUPPORT)
class PerformDragControllerAction {
public:
    using Arguments = std::tuple<WebKit::DragControllerAction, const WebCore::IntPoint&, const WebCore::IntPoint&, const OptionSet<WebCore::DragOperation>&, const WebCore::SelectionData&, const OptionSet<WebCore::DragApplicationFlags>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_PerformDragControllerAction; }
    static const bool isSync = false;

    PerformDragControllerAction(WebKit::DragControllerAction action, const WebCore::IntPoint& clientPosition, const WebCore::IntPoint& globalPosition, const OptionSet<WebCore::DragOperation>& draggingSourceOperationMask, const WebCore::SelectionData& selection, const OptionSet<WebCore::DragApplicationFlags>& flags)
        : m_arguments(action, clientPosition, globalPosition, draggingSourceOperationMask, selection, flags)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if !PLATFORM(GTK) && ENABLE(DRAG_SUPPORT)
class PerformDragControllerAction {
public:
    using Arguments = std::tuple<WebKit::DragControllerAction, const WebCore::DragData&, const WebKit::SandboxExtension::Handle&, const WebKit::SandboxExtension::HandleArray&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_PerformDragControllerAction; }
    static const bool isSync = false;

    PerformDragControllerAction(WebKit::DragControllerAction action, const WebCore::DragData& dragData, const WebKit::SandboxExtension::Handle& sandboxExtensionHandle, const WebKit::SandboxExtension::HandleArray& sandboxExtensionsForUpload)
        : m_arguments(action, dragData, sandboxExtensionHandle, sandboxExtensionsForUpload)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(DRAG_SUPPORT)
class DidStartDrag {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidStartDrag; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(DRAG_SUPPORT)
class DragEnded {
public:
    using Arguments = std::tuple<const WebCore::IntPoint&, const WebCore::IntPoint&, const OptionSet<WebCore::DragOperation>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DragEnded; }
    static const bool isSync = false;

    DragEnded(const WebCore::IntPoint& clientPosition, const WebCore::IntPoint& globalPosition, const OptionSet<WebCore::DragOperation>& dragOperationMask)
        : m_arguments(clientPosition, globalPosition, dragOperationMask)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(DRAG_SUPPORT)
class DragCancelled {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DragCancelled; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY) && ENABLE(DRAG_SUPPORT)
class RequestDragStart {
public:
    using Arguments = std::tuple<const WebCore::IntPoint&, const WebCore::IntPoint&, const OptionSet<WebCore::DragSourceAction>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RequestDragStart; }
    static const bool isSync = false;

    RequestDragStart(const WebCore::IntPoint& clientPosition, const WebCore::IntPoint& globalPosition, const OptionSet<WebCore::DragSourceAction>& allowedActionsMask)
        : m_arguments(clientPosition, globalPosition, allowedActionsMask)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY) && ENABLE(DRAG_SUPPORT)
class RequestAdditionalItemsForDragSession {
public:
    using Arguments = std::tuple<const WebCore::IntPoint&, const WebCore::IntPoint&, const OptionSet<WebCore::DragSourceAction>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RequestAdditionalItemsForDragSession; }
    static const bool isSync = false;

    RequestAdditionalItemsForDragSession(const WebCore::IntPoint& clientPosition, const WebCore::IntPoint& globalPosition, const OptionSet<WebCore::DragSourceAction>& allowedActionsMask)
        : m_arguments(clientPosition, globalPosition, allowedActionsMask)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY) && ENABLE(DRAG_SUPPORT)
class InsertDroppedImagePlaceholders {
public:
    using Arguments = std::tuple<const Vector<WebCore::IntSize>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_InsertDroppedImagePlaceholders; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Vector<WebCore::IntRect>&&, Optional<WebCore::TextIndicatorData>&&)>&&);
    static void cancelReply(CompletionHandler<void(Vector<WebCore::IntRect>&&, Optional<WebCore::TextIndicatorData>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_InsertDroppedImagePlaceholdersReply; }
    using AsyncReply = InsertDroppedImagePlaceholdersAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<WebCore::IntRect>& dropPreviewRects, const Optional<WebCore::TextIndicatorData>& textIndicator);
    using Reply = std::tuple<Vector<WebCore::IntRect>&, Optional<WebCore::TextIndicatorData>&>;
    using ReplyArguments = std::tuple<Vector<WebCore::IntRect>, Optional<WebCore::TextIndicatorData>>;
    explicit InsertDroppedImagePlaceholders(const Vector<WebCore::IntSize>& imageSize)
        : m_arguments(imageSize)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY) && ENABLE(DRAG_SUPPORT)
class DidConcludeDrop {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidConcludeDrop; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class DidChangeSelectedIndexForActivePopupMenu {
public:
    using Arguments = std::tuple<int32_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidChangeSelectedIndexForActivePopupMenu; }
    static const bool isSync = false;

    explicit DidChangeSelectedIndexForActivePopupMenu(int32_t newIndex)
        : m_arguments(newIndex)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetTextForActivePopupMenu {
public:
    using Arguments = std::tuple<int32_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetTextForActivePopupMenu; }
    static const bool isSync = false;

    explicit SetTextForActivePopupMenu(int32_t index)
        : m_arguments(index)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(GTK)
class FailedToShowPopupMenu {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_FailedToShowPopupMenu; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(CONTEXT_MENUS)
class DidSelectItemFromActiveContextMenu {
public:
    using Arguments = std::tuple<const WebKit::WebContextMenuItemData&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidSelectItemFromActiveContextMenu; }
    static const bool isSync = false;

    explicit DidSelectItemFromActiveContextMenu(const WebKit::WebContextMenuItemData& menuItem)
        : m_arguments(menuItem)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class DidChooseFilesForOpenPanelWithDisplayStringAndIcon {
public:
    using Arguments = std::tuple<const Vector<String>&, const String&, const IPC::DataReference&, const WebKit::SandboxExtension::Handle&, const WebKit::SandboxExtension::Handle&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidChooseFilesForOpenPanelWithDisplayStringAndIcon; }
    static const bool isSync = false;

    DidChooseFilesForOpenPanelWithDisplayStringAndIcon(const Vector<String>& fileURLs, const String& displayString, const IPC::DataReference& iconData, const WebKit::SandboxExtension::Handle& frontboardServicesSandboxExtension, const WebKit::SandboxExtension::Handle& iconServicesSandboxExtension)
        : m_arguments(fileURLs, displayString, iconData, frontboardServicesSandboxExtension, iconServicesSandboxExtension)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class DidChooseFilesForOpenPanel {
public:
    using Arguments = std::tuple<const Vector<String>&, const Vector<String>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidChooseFilesForOpenPanel; }
    static const bool isSync = false;

    DidChooseFilesForOpenPanel(const Vector<String>& fileURLs, const Vector<String>& replacementURLs)
        : m_arguments(fileURLs, replacementURLs)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidCancelForOpenPanel {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidCancelForOpenPanel; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if ENABLE(SANDBOX_EXTENSIONS)
class ExtendSandboxForFilesFromOpenPanel {
public:
    using Arguments = std::tuple<const WebKit::SandboxExtension::HandleArray&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ExtendSandboxForFilesFromOpenPanel; }
    static const bool isSync = false;

    explicit ExtendSandboxForFilesFromOpenPanel(const WebKit::SandboxExtension::HandleArray& sandboxExtensions)
        : m_arguments(sandboxExtensions)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class AdvanceToNextMisspelling {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_AdvanceToNextMisspelling; }
    static const bool isSync = false;

    explicit AdvanceToNextMisspelling(bool startBeforeSelection)
        : m_arguments(startBeforeSelection)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ChangeSpellingToWord {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ChangeSpellingToWord; }
    static const bool isSync = false;

    explicit ChangeSpellingToWord(const String& word)
        : m_arguments(word)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidFinishCheckingText {
public:
    using Arguments = std::tuple<uint64_t, const Vector<WebCore::TextCheckingResult>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidFinishCheckingText; }
    static const bool isSync = false;

    DidFinishCheckingText(uint64_t requestID, const Vector<WebCore::TextCheckingResult>& result)
        : m_arguments(requestID, result)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidCancelCheckingText {
public:
    using Arguments = std::tuple<uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidCancelCheckingText; }
    static const bool isSync = false;

    explicit DidCancelCheckingText(uint64_t requestID)
        : m_arguments(requestID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if USE(APPKIT)
class UppercaseWord {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_UppercaseWord; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if USE(APPKIT)
class LowercaseWord {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_LowercaseWord; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if USE(APPKIT)
class CapitalizeWord {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_CapitalizeWord; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetSmartInsertDeleteEnabled {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetSmartInsertDeleteEnabled; }
    static const bool isSync = false;

    explicit SetSmartInsertDeleteEnabled(bool isSmartInsertDeleteEnabled)
        : m_arguments(isSmartInsertDeleteEnabled)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(GEOLOCATION)
class DidReceiveGeolocationPermissionDecision {
public:
    using Arguments = std::tuple<const WebKit::GeolocationIdentifier&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidReceiveGeolocationPermissionDecision; }
    static const bool isSync = false;

    DidReceiveGeolocationPermissionDecision(const WebKit::GeolocationIdentifier& geolocationID, const String& authorizationToken)
        : m_arguments(geolocationID, authorizationToken)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(MEDIA_STREAM)
class UserMediaAccessWasGranted {
public:
    using Arguments = std::tuple<uint64_t, const WebCore::CaptureDevice&, const WebCore::CaptureDevice&, const String&, const WebKit::SandboxExtension::Handle&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_UserMediaAccessWasGranted; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_UserMediaAccessWasGrantedReply; }
    using AsyncReply = UserMediaAccessWasGrantedAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    UserMediaAccessWasGranted(uint64_t userMediaID, const WebCore::CaptureDevice& audioDevice, const WebCore::CaptureDevice& videoDevice, const String& mediaDeviceIdentifierHashSalt, const WebKit::SandboxExtension::Handle& sandboxExtensionHandle)
        : m_arguments(userMediaID, audioDevice, videoDevice, mediaDeviceIdentifierHashSalt, sandboxExtensionHandle)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(MEDIA_STREAM)
class UserMediaAccessWasDenied {
public:
    using Arguments = std::tuple<uint64_t, uint64_t, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_UserMediaAccessWasDenied; }
    static const bool isSync = false;

    UserMediaAccessWasDenied(uint64_t userMediaID, uint64_t reason, const String& invalidConstraint)
        : m_arguments(userMediaID, reason, invalidConstraint)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(MEDIA_STREAM)
class CaptureDevicesChanged {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_CaptureDevicesChanged; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(ENCRYPTED_MEDIA)
class MediaKeySystemWasGranted {
public:
    using Arguments = std::tuple<uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_MediaKeySystemWasGranted; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_MediaKeySystemWasGrantedReply; }
    using AsyncReply = MediaKeySystemWasGrantedAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    explicit MediaKeySystemWasGranted(uint64_t mediaKeySystemID)
        : m_arguments(mediaKeySystemID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(ENCRYPTED_MEDIA)
class MediaKeySystemWasDenied {
public:
    using Arguments = std::tuple<uint64_t, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_MediaKeySystemWasDenied; }
    static const bool isSync = false;

    MediaKeySystemWasDenied(uint64_t mediaKeySystemID, const String& message)
        : m_arguments(mediaKeySystemID, message)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class RequestMediaPlaybackState {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RequestMediaPlaybackState; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebKit::MediaPlaybackState&&)>&&);
    static void cancelReply(CompletionHandler<void(WebKit::MediaPlaybackState&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_RequestMediaPlaybackStateReply; }
    using AsyncReply = RequestMediaPlaybackStateAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, WebKit::MediaPlaybackState state);
    using Reply = std::tuple<WebKit::MediaPlaybackState&>;
    using ReplyArguments = std::tuple<WebKit::MediaPlaybackState>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class PauseAllMediaPlayback {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_PauseAllMediaPlayback; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_PauseAllMediaPlaybackReply; }
    using AsyncReply = PauseAllMediaPlaybackAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SuspendAllMediaPlayback {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SuspendAllMediaPlayback; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_SuspendAllMediaPlaybackReply; }
    using AsyncReply = SuspendAllMediaPlaybackAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ResumeAllMediaPlayback {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ResumeAllMediaPlayback; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_ResumeAllMediaPlaybackReply; }
    using AsyncReply = ResumeAllMediaPlaybackAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidReceiveNotificationPermissionDecision {
public:
    using Arguments = std::tuple<uint64_t, bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidReceiveNotificationPermissionDecision; }
    static const bool isSync = false;

    DidReceiveNotificationPermissionDecision(uint64_t notificationID, bool allowed)
        : m_arguments(notificationID, allowed)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class FreezeLayerTreeDueToSwipeAnimation {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_FreezeLayerTreeDueToSwipeAnimation; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class UnfreezeLayerTreeDueToSwipeAnimation {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_UnfreezeLayerTreeDueToSwipeAnimation; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class BeginPrinting {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&, const WebKit::PrintInfo&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_BeginPrinting; }
    static const bool isSync = false;

    BeginPrinting(const WebCore::FrameIdentifier& frameID, const WebKit::PrintInfo& printInfo)
        : m_arguments(frameID, printInfo)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class EndPrinting {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_EndPrinting; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ComputePagesForPrinting {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&, const WebKit::PrintInfo&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ComputePagesForPrinting; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Vector<WebCore::IntRect>&&, double&&, WebCore::RectEdges<float>&&)>&&);
    static void cancelReply(CompletionHandler<void(Vector<WebCore::IntRect>&&, double&&, WebCore::RectEdges<float>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_ComputePagesForPrintingReply; }
    using AsyncReply = ComputePagesForPrintingAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<WebCore::IntRect>& pageRects, double totalScaleFactorForPrinting, const WebCore::RectEdges<float>& computedPageMargin);
    using Reply = std::tuple<Vector<WebCore::IntRect>&, double&, WebCore::RectEdges<float>&>;
    using ReplyArguments = std::tuple<Vector<WebCore::IntRect>, double, WebCore::RectEdges<float>>;
    ComputePagesForPrinting(const WebCore::FrameIdentifier& frameID, const WebKit::PrintInfo& printInfo)
        : m_arguments(frameID, printInfo)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(COCOA)
class DrawRectToImage {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&, const WebKit::PrintInfo&, const WebCore::IntRect&, const WebCore::IntSize&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DrawRectToImage; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebKit::ShareableBitmap::Handle&&)>&&);
    static void cancelReply(CompletionHandler<void(WebKit::ShareableBitmap::Handle&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_DrawRectToImageReply; }
    using AsyncReply = DrawRectToImageAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::ShareableBitmap::Handle& image);
    using Reply = std::tuple<WebKit::ShareableBitmap::Handle&>;
    using ReplyArguments = std::tuple<WebKit::ShareableBitmap::Handle>;
    DrawRectToImage(const WebCore::FrameIdentifier& frameID, const WebKit::PrintInfo& printInfo, const WebCore::IntRect& rect, const WebCore::IntSize& imageSize)
        : m_arguments(frameID, printInfo, rect, imageSize)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class DrawPagesToPDF {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&, const WebKit::PrintInfo&, uint32_t, uint32_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DrawPagesToPDF; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(IPC::DataReference&&)>&&);
    static void cancelReply(CompletionHandler<void(IPC::DataReference&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_DrawPagesToPDFReply; }
    using AsyncReply = DrawPagesToPDFAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const IPC::DataReference& data);
    using Reply = std::tuple<IPC::DataReference&>;
    using ReplyArguments = std::tuple<IPC::DataReference>;
    DrawPagesToPDF(const WebCore::FrameIdentifier& frameID, const WebKit::PrintInfo& printInfo, uint32_t first, uint32_t count)
        : m_arguments(frameID, printInfo, first, count)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if (PLATFORM(COCOA) && PLATFORM(IOS_FAMILY))
class ComputePagesForPrintingiOS {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&, const WebKit::PrintInfo&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ComputePagesForPrintingiOS; }
    static const bool isSync = true;

    using DelayedReply = ComputePagesForPrintingiOSDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, size_t pageCount);
    using Reply = std::tuple<size_t&>;
    using ReplyArguments = std::tuple<size_t>;
    ComputePagesForPrintingiOS(const WebCore::FrameIdentifier& frameID, const WebKit::PrintInfo& printInfo)
        : m_arguments(frameID, printInfo)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if (PLATFORM(COCOA) && PLATFORM(IOS_FAMILY))
class DrawToPDFiOS {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&, const WebKit::PrintInfo&, size_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DrawToPDFiOS; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(IPC::DataReference&&)>&&);
    static void cancelReply(CompletionHandler<void(IPC::DataReference&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_DrawToPDFiOSReply; }
    using AsyncReply = DrawToPDFiOSAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const IPC::DataReference& data);
    using Reply = std::tuple<IPC::DataReference&>;
    using ReplyArguments = std::tuple<IPC::DataReference>;
    DrawToPDFiOS(const WebCore::FrameIdentifier& frameID, const WebKit::PrintInfo& printInfo, size_t pageCount)
        : m_arguments(frameID, printInfo, pageCount)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class DrawToPDF {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&, const Optional<WebCore::FloatRect>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DrawToPDF; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(IPC::DataReference&&)>&&);
    static void cancelReply(CompletionHandler<void(IPC::DataReference&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_DrawToPDFReply; }
    using AsyncReply = DrawToPDFAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const IPC::DataReference& data);
    using Reply = std::tuple<IPC::DataReference&>;
    using ReplyArguments = std::tuple<IPC::DataReference>;
    DrawToPDF(const WebCore::FrameIdentifier& frameID, const Optional<WebCore::FloatRect>& rect)
        : m_arguments(frameID, rect)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(GTK)
class DrawPagesForPrinting {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&, const WebKit::PrintInfo&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DrawPagesForPrinting; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebCore::ResourceError&&)>&&);
    static void cancelReply(CompletionHandler<void(WebCore::ResourceError&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_DrawPagesForPrintingReply; }
    using AsyncReply = DrawPagesForPrintingAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebCore::ResourceError& error);
    using Reply = std::tuple<WebCore::ResourceError&>;
    using ReplyArguments = std::tuple<WebCore::ResourceError>;
    DrawPagesForPrinting(const WebCore::FrameIdentifier& frameID, const WebKit::PrintInfo& printInfo)
        : m_arguments(frameID, printInfo)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class SetMediaVolume {
public:
    using Arguments = std::tuple<float>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetMediaVolume; }
    static const bool isSync = false;

    explicit SetMediaVolume(float volume)
        : m_arguments(volume)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetMuted {
public:
    using Arguments = std::tuple<const WebCore::MediaProducer::MutedStateFlags&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetMuted; }
    static const bool isSync = false;

    explicit SetMuted(const WebCore::MediaProducer::MutedStateFlags& muted)
        : m_arguments(muted)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetMayStartMediaWhenInWindow {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetMayStartMediaWhenInWindow; }
    static const bool isSync = false;

    explicit SetMayStartMediaWhenInWindow(bool mayStartMedia)
        : m_arguments(mayStartMedia)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class StopMediaCapture {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_StopMediaCapture; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetCanRunBeforeUnloadConfirmPanel {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetCanRunBeforeUnloadConfirmPanel; }
    static const bool isSync = false;

    explicit SetCanRunBeforeUnloadConfirmPanel(bool canRunBeforeUnloadConfirmPanel)
        : m_arguments(canRunBeforeUnloadConfirmPanel)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetCanRunModal {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetCanRunModal; }
    static const bool isSync = false;

    explicit SetCanRunModal(bool canRunModal)
        : m_arguments(canRunModal)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(GTK) || PLATFORM(WPE)
class CancelComposition {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_CancelComposition; }
    static const bool isSync = false;

    explicit CancelComposition(const String& text)
        : m_arguments(text)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(GTK) || PLATFORM(WPE)
class DeleteSurrounding {
public:
    using Arguments = std::tuple<int64_t, const unsigned&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DeleteSurrounding; }
    static const bool isSync = false;

    DeleteSurrounding(int64_t offset, const unsigned& characterCount)
        : m_arguments(offset, characterCount)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(GTK)
class CollapseSelectionInFrame {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_CollapseSelectionInFrame; }
    static const bool isSync = false;

    explicit CollapseSelectionInFrame(const WebCore::FrameIdentifier& frameID)
        : m_arguments(frameID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(GTK)
class GetCenterForZoomGesture {
public:
    using Arguments = std::tuple<const WebCore::IntPoint&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetCenterForZoomGesture; }
    static const bool isSync = true;

    using DelayedReply = GetCenterForZoomGestureDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebCore::IntPoint& center);
    using Reply = std::tuple<WebCore::IntPoint&>;
    using ReplyArguments = std::tuple<WebCore::IntPoint>;
    explicit GetCenterForZoomGesture(const WebCore::IntPoint& centerInViewCoordinates)
        : m_arguments(centerInViewCoordinates)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SendComplexTextInputToPlugin {
public:
    using Arguments = std::tuple<uint64_t, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SendComplexTextInputToPlugin; }
    static const bool isSync = false;

    SendComplexTextInputToPlugin(uint64_t pluginComplexTextInputIdentifier, const String& textInput)
        : m_arguments(pluginComplexTextInputIdentifier, textInput)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class WindowAndViewFramesChanged {
public:
    using Arguments = std::tuple<const WebCore::FloatRect&, const WebCore::FloatRect&, const WebCore::FloatRect&, const WebCore::FloatPoint&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_WindowAndViewFramesChanged; }
    static const bool isSync = false;

    WindowAndViewFramesChanged(const WebCore::FloatRect& windowFrameInScreenCoordinates, const WebCore::FloatRect& windowFrameInUnflippedScreenCoordinates, const WebCore::FloatRect& viewFrameInWindowCoordinates, const WebCore::FloatPoint& accessibilityViewCoordinates)
        : m_arguments(windowFrameInScreenCoordinates, windowFrameInUnflippedScreenCoordinates, viewFrameInWindowCoordinates, accessibilityViewCoordinates)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetMainFrameIsScrollable {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetMainFrameIsScrollable; }
    static const bool isSync = false;

    explicit SetMainFrameIsScrollable(bool isScrollable)
        : m_arguments(isScrollable)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class RegisterUIProcessAccessibilityTokens {
public:
    using Arguments = std::tuple<const IPC::DataReference&, const IPC::DataReference&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RegisterUIProcessAccessibilityTokens; }
    static const bool isSync = false;

    RegisterUIProcessAccessibilityTokens(const IPC::DataReference& elemenToken, const IPC::DataReference& windowToken)
        : m_arguments(elemenToken, windowToken)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetStringSelectionForPasteboard {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetStringSelectionForPasteboard; }
    static const bool isSync = true;

    using DelayedReply = GetStringSelectionForPasteboardDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& stringValue);
    using Reply = std::tuple<String&>;
    using ReplyArguments = std::tuple<String>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetDataSelectionForPasteboard {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetDataSelectionForPasteboard; }
    static const bool isSync = true;

    using DelayedReply = GetDataSelectionForPasteboardDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::SharedMemory::IPCHandle& ipcHandle);
    using Reply = std::tuple<WebKit::SharedMemory::IPCHandle&>;
    using ReplyArguments = std::tuple<WebKit::SharedMemory::IPCHandle>;
    explicit GetDataSelectionForPasteboard(const String& pasteboardType)
        : m_arguments(pasteboardType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class ReadSelectionFromPasteboard {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ReadSelectionFromPasteboard; }
    static const bool isSync = true;

    using DelayedReply = ReadSelectionFromPasteboardDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool result);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit ReadSelectionFromPasteboard(const String& pasteboardName)
        : m_arguments(pasteboardName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if (PLATFORM(COCOA) && ENABLE(SERVICE_CONTROLS))
class ReplaceSelectionWithPasteboardData {
public:
    using Arguments = std::tuple<const Vector<String>&, const IPC::DataReference&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ReplaceSelectionWithPasteboardData; }
    static const bool isSync = false;

    ReplaceSelectionWithPasteboardData(const Vector<String>& types, const IPC::DataReference& data)
        : m_arguments(types, data)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class ShouldDelayWindowOrderingEvent {
public:
    using Arguments = std::tuple<const WebKit::WebMouseEvent&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ShouldDelayWindowOrderingEvent; }
    static const bool isSync = true;

    using DelayedReply = ShouldDelayWindowOrderingEventDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool result);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit ShouldDelayWindowOrderingEvent(const WebKit::WebMouseEvent& event)
        : m_arguments(event)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class AcceptsFirstMouse {
public:
    using Arguments = std::tuple<const int&, const WebKit::WebMouseEvent&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_AcceptsFirstMouse; }
    static const bool isSync = true;

    using DelayedReply = AcceptsFirstMouseDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool result);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    AcceptsFirstMouse(const int& eventNumber, const WebKit::WebMouseEvent& event)
        : m_arguments(eventNumber, event)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetTextAsync {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetTextAsync; }
    static const bool isSync = false;

    explicit SetTextAsync(const String& text)
        : m_arguments(text)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class InsertTextAsync {
public:
    using Arguments = std::tuple<const String&, const WebKit::EditingRange&, const WebKit::InsertTextOptions&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_InsertTextAsync; }
    static const bool isSync = false;

    InsertTextAsync(const String& text, const WebKit::EditingRange& replacementRange, const WebKit::InsertTextOptions& options)
        : m_arguments(text, replacementRange, options)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class InsertDictatedTextAsync {
public:
    using Arguments = std::tuple<const String&, const WebKit::EditingRange&, const Vector<WebCore::DictationAlternative>&, const WebKit::InsertTextOptions&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_InsertDictatedTextAsync; }
    static const bool isSync = false;

    InsertDictatedTextAsync(const String& text, const WebKit::EditingRange& replacementRange, const Vector<WebCore::DictationAlternative>& dictationAlternatives, const WebKit::InsertTextOptions& options)
        : m_arguments(text, replacementRange, dictationAlternatives, options)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class HasMarkedText {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_HasMarkedText; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_HasMarkedTextReply; }
    using AsyncReply = HasMarkedTextAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool hasMarkedText);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetMarkedRangeAsync {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetMarkedRangeAsync; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebKit::EditingRange&&)>&&);
    static void cancelReply(CompletionHandler<void(WebKit::EditingRange&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetMarkedRangeAsyncReply; }
    using AsyncReply = GetMarkedRangeAsyncAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::EditingRange& range);
    using Reply = std::tuple<WebKit::EditingRange&>;
    using ReplyArguments = std::tuple<WebKit::EditingRange>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class GetSelectedRangeAsync {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetSelectedRangeAsync; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebKit::EditingRange&&)>&&);
    static void cancelReply(CompletionHandler<void(WebKit::EditingRange&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetSelectedRangeAsyncReply; }
    using AsyncReply = GetSelectedRangeAsyncAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::EditingRange& range);
    using Reply = std::tuple<WebKit::EditingRange&>;
    using ReplyArguments = std::tuple<WebKit::EditingRange>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class CharacterIndexForPointAsync {
public:
    using Arguments = std::tuple<const WebCore::IntPoint&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_CharacterIndexForPointAsync; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(uint64_t&&)>&&);
    static void cancelReply(CompletionHandler<void(uint64_t&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_CharacterIndexForPointAsyncReply; }
    using AsyncReply = CharacterIndexForPointAsyncAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, uint64_t location);
    using Reply = std::tuple<uint64_t&>;
    using ReplyArguments = std::tuple<uint64_t>;
    explicit CharacterIndexForPointAsync(const WebCore::IntPoint& point)
        : m_arguments(point)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class FirstRectForCharacterRangeAsync {
public:
    using Arguments = std::tuple<const WebKit::EditingRange&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_FirstRectForCharacterRangeAsync; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebCore::IntRect&&, WebKit::EditingRange&&)>&&);
    static void cancelReply(CompletionHandler<void(WebCore::IntRect&&, WebKit::EditingRange&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_FirstRectForCharacterRangeAsyncReply; }
    using AsyncReply = FirstRectForCharacterRangeAsyncAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebCore::IntRect& rect, const WebKit::EditingRange& actualRange);
    using Reply = std::tuple<WebCore::IntRect&, WebKit::EditingRange&>;
    using ReplyArguments = std::tuple<WebCore::IntRect, WebKit::EditingRange>;
    explicit FirstRectForCharacterRangeAsync(const WebKit::EditingRange& range)
        : m_arguments(range)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class SetCompositionAsync {
public:
    using Arguments = std::tuple<const String&, const Vector<WebCore::CompositionUnderline>&, const Vector<WebCore::CompositionHighlight>&, const WebKit::EditingRange&, const WebKit::EditingRange&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetCompositionAsync; }
    static const bool isSync = false;

    SetCompositionAsync(const String& text, const Vector<WebCore::CompositionUnderline>& underlines, const Vector<WebCore::CompositionHighlight>& highlights, const WebKit::EditingRange& selectionRange, const WebKit::EditingRange& replacementRange)
        : m_arguments(text, underlines, highlights, selectionRange, replacementRange)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class ConfirmCompositionAsync {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ConfirmCompositionAsync; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(MAC)
class AttributedSubstringForCharacterRangeAsync {
public:
    using Arguments = std::tuple<const WebKit::EditingRange&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_AttributedSubstringForCharacterRangeAsync; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebCore::AttributedString&&, WebKit::EditingRange&&)>&&);
    static void cancelReply(CompletionHandler<void(WebCore::AttributedString&&, WebKit::EditingRange&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_AttributedSubstringForCharacterRangeAsyncReply; }
    using AsyncReply = AttributedSubstringForCharacterRangeAsyncAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebCore::AttributedString& string, const WebKit::EditingRange& range);
    using Reply = std::tuple<WebCore::AttributedString&, WebKit::EditingRange&>;
    using ReplyArguments = std::tuple<WebCore::AttributedString, WebKit::EditingRange>;
    explicit AttributedSubstringForCharacterRangeAsync(const WebKit::EditingRange& range)
        : m_arguments(range)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(MAC)
class FontAtSelection {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_FontAtSelection; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebKit::FontInfo&&, double&&, bool&&)>&&);
    static void cancelReply(CompletionHandler<void(WebKit::FontInfo&&, double&&, bool&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_FontAtSelectionReply; }
    using AsyncReply = FontAtSelectionAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::FontInfo& fontInfo, double fontSize, bool selectionHasMultipleFonts);
    using Reply = std::tuple<WebKit::FontInfo&, double&, bool&>;
    using ReplyArguments = std::tuple<WebKit::FontInfo, double, bool>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class SetAlwaysShowsHorizontalScroller {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetAlwaysShowsHorizontalScroller; }
    static const bool isSync = false;

    explicit SetAlwaysShowsHorizontalScroller(bool alwaysShowsHorizontalScroller)
        : m_arguments(alwaysShowsHorizontalScroller)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetAlwaysShowsVerticalScroller {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetAlwaysShowsVerticalScroller; }
    static const bool isSync = false;

    explicit SetAlwaysShowsVerticalScroller(bool alwaysShowsVerticalScroller)
        : m_arguments(alwaysShowsVerticalScroller)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetMinimumSizeForAutoLayout {
public:
    using Arguments = std::tuple<const WebCore::IntSize&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetMinimumSizeForAutoLayout; }
    static const bool isSync = false;

    explicit SetMinimumSizeForAutoLayout(const WebCore::IntSize& size)
        : m_arguments(size)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetSizeToContentAutoSizeMaximumSize {
public:
    using Arguments = std::tuple<const WebCore::IntSize&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetSizeToContentAutoSizeMaximumSize; }
    static const bool isSync = false;

    explicit SetSizeToContentAutoSizeMaximumSize(const WebCore::IntSize& size)
        : m_arguments(size)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetAutoSizingShouldExpandToViewHeight {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetAutoSizingShouldExpandToViewHeight; }
    static const bool isSync = false;

    explicit SetAutoSizingShouldExpandToViewHeight(bool shouldExpand)
        : m_arguments(shouldExpand)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetViewportSizeForCSSViewportUnits {
public:
    using Arguments = std::tuple<const Optional<WebCore::IntSize>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetViewportSizeForCSSViewportUnits; }
    static const bool isSync = false;

    explicit SetViewportSizeForCSSViewportUnits(const Optional<WebCore::IntSize>& viewportSize)
        : m_arguments(viewportSize)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(COCOA)
class HandleAlternativeTextUIResult {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_HandleAlternativeTextUIResult; }
    static const bool isSync = false;

    explicit HandleAlternativeTextUIResult(const String& result)
        : m_arguments(result)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(IOS_FAMILY)
class WillStartUserTriggeredZooming {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_WillStartUserTriggeredZooming; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class SetScrollPinningBehavior {
public:
    using Arguments = std::tuple<uint32_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetScrollPinningBehavior; }
    static const bool isSync = false;

    explicit SetScrollPinningBehavior(uint32_t pinning)
        : m_arguments(pinning)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetScrollbarOverlayStyle {
public:
    using Arguments = std::tuple<const Optional<uint32_t>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetScrollbarOverlayStyle; }
    static const bool isSync = false;

    explicit SetScrollbarOverlayStyle(const Optional<uint32_t>& scrollbarStyle)
        : m_arguments(scrollbarStyle)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetBytecodeProfile {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetBytecodeProfile; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(String&&)>&&);
    static void cancelReply(CompletionHandler<void(String&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetBytecodeProfileReply; }
    using AsyncReply = GetBytecodeProfileAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& string);
    using Reply = std::tuple<String&>;
    using ReplyArguments = std::tuple<String>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetSamplingProfilerOutput {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetSamplingProfilerOutput; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(String&&)>&&);
    static void cancelReply(CompletionHandler<void(String&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetSamplingProfilerOutputReply; }
    using AsyncReply = GetSamplingProfilerOutputAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& string);
    using Reply = std::tuple<String&>;
    using ReplyArguments = std::tuple<String>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class TakeSnapshot {
public:
    using Arguments = std::tuple<const WebCore::IntRect&, const WebCore::IntSize&, uint32_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_TakeSnapshot; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebKit::ShareableBitmap::Handle&&)>&&);
    static void cancelReply(CompletionHandler<void(WebKit::ShareableBitmap::Handle&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_TakeSnapshotReply; }
    using AsyncReply = TakeSnapshotAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::ShareableBitmap::Handle& image);
    using Reply = std::tuple<WebKit::ShareableBitmap::Handle&>;
    using ReplyArguments = std::tuple<WebKit::ShareableBitmap::Handle>;
    TakeSnapshot(const WebCore::IntRect& snapshotRect, const WebCore::IntSize& bitmapSize, uint32_t options)
        : m_arguments(snapshotRect, bitmapSize, options)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(MAC)
class PerformImmediateActionHitTestAtLocation {
public:
    using Arguments = std::tuple<const WebCore::FloatPoint&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_PerformImmediateActionHitTestAtLocation; }
    static const bool isSync = false;

    explicit PerformImmediateActionHitTestAtLocation(const WebCore::FloatPoint& location)
        : m_arguments(location)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(MAC)
class ImmediateActionDidUpdate {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ImmediateActionDidUpdate; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(MAC)
class ImmediateActionDidCancel {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ImmediateActionDidCancel; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(MAC)
class ImmediateActionDidComplete {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ImmediateActionDidComplete; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(MAC)
class DataDetectorsDidPresentUI {
public:
    using Arguments = std::tuple<const WebCore::PageOverlay::PageOverlayID&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DataDetectorsDidPresentUI; }
    static const bool isSync = false;

    explicit DataDetectorsDidPresentUI(const WebCore::PageOverlay::PageOverlayID& pageOverlay)
        : m_arguments(pageOverlay)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(MAC)
class DataDetectorsDidChangeUI {
public:
    using Arguments = std::tuple<const WebCore::PageOverlay::PageOverlayID&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DataDetectorsDidChangeUI; }
    static const bool isSync = false;

    explicit DataDetectorsDidChangeUI(const WebCore::PageOverlay::PageOverlayID& pageOverlay)
        : m_arguments(pageOverlay)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(MAC)
class DataDetectorsDidHideUI {
public:
    using Arguments = std::tuple<const WebCore::PageOverlay::PageOverlayID&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DataDetectorsDidHideUI; }
    static const bool isSync = false;

    explicit DataDetectorsDidHideUI(const WebCore::PageOverlay::PageOverlayID& pageOverlay)
        : m_arguments(pageOverlay)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(MAC)
class HandleAcceptedCandidate {
public:
    using Arguments = std::tuple<const WebCore::TextCheckingResult&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_HandleAcceptedCandidate; }
    static const bool isSync = false;

    explicit HandleAcceptedCandidate(const WebCore::TextCheckingResult& acceptedCandidate)
        : m_arguments(acceptedCandidate)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(MAC)
class SetUseSystemAppearance {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetUseSystemAppearance; }
    static const bool isSync = false;

    explicit SetUseSystemAppearance(bool useSystemAppearance)
        : m_arguments(useSystemAppearance)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(MAC)
class SetHeaderBannerHeightForTesting {
public:
    using Arguments = std::tuple<const int&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetHeaderBannerHeightForTesting; }
    static const bool isSync = false;

    explicit SetHeaderBannerHeightForTesting(const int& height)
        : m_arguments(height)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(MAC)
class SetFooterBannerHeightForTesting {
public:
    using Arguments = std::tuple<const int&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetFooterBannerHeightForTesting; }
    static const bool isSync = false;

    explicit SetFooterBannerHeightForTesting(const int& height)
        : m_arguments(height)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(MAC)
class DidEndMagnificationGesture {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidEndMagnificationGesture; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class EffectiveAppearanceDidChange {
public:
    using Arguments = std::tuple<bool, bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_EffectiveAppearanceDidChange; }
    static const bool isSync = false;

    EffectiveAppearanceDidChange(bool useDarkAppearance, bool useElevatedUserInterfaceLevel)
        : m_arguments(useDarkAppearance, useElevatedUserInterfaceLevel)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(GTK)
class ThemeDidChange {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ThemeDidChange; }
    static const bool isSync = false;

    explicit ThemeDidChange(const String& themeName)
        : m_arguments(themeName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(COCOA)
class RequestActiveNowPlayingSessionInfo {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RequestActiveNowPlayingSessionInfo; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&, bool&&, String&&, double&&, double&&, uint64_t&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&, bool&&, String&&, double&&, double&&, uint64_t&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_RequestActiveNowPlayingSessionInfoReply; }
    using AsyncReply = RequestActiveNowPlayingSessionInfoAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool active, bool registeredAsNowPlayingApplication, const String& title, double duration, double elapsedTime, uint64_t uniqueIdentifier);
    using Reply = std::tuple<bool&, bool&, String&, double&, double&, uint64_t&>;
    using ReplyArguments = std::tuple<bool, bool, String, double, double, uint64_t>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class SetShouldDispatchFakeMouseMoveEvents {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetShouldDispatchFakeMouseMoveEvents; }
    static const bool isSync = false;

    explicit SetShouldDispatchFakeMouseMoveEvents(bool shouldDispatchFakeMouseMoveEvents)
        : m_arguments(shouldDispatchFakeMouseMoveEvents)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
class PlaybackTargetSelected {
public:
    using Arguments = std::tuple<const WebCore::PlaybackTargetClientContextIdentifier&, const WebCore::MediaPlaybackTargetContext&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_PlaybackTargetSelected; }
    static const bool isSync = false;

    PlaybackTargetSelected(const WebCore::PlaybackTargetClientContextIdentifier& contextId, const WebCore::MediaPlaybackTargetContext& target)
        : m_arguments(contextId, target)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
class PlaybackTargetAvailabilityDidChange {
public:
    using Arguments = std::tuple<const WebCore::PlaybackTargetClientContextIdentifier&, bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_PlaybackTargetAvailabilityDidChange; }
    static const bool isSync = false;

    PlaybackTargetAvailabilityDidChange(const WebCore::PlaybackTargetClientContextIdentifier& contextId, bool available)
        : m_arguments(contextId, available)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
class SetShouldPlayToPlaybackTarget {
public:
    using Arguments = std::tuple<const WebCore::PlaybackTargetClientContextIdentifier&, bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetShouldPlayToPlaybackTarget; }
    static const bool isSync = false;

    SetShouldPlayToPlaybackTarget(const WebCore::PlaybackTargetClientContextIdentifier& contextId, bool shouldPlay)
        : m_arguments(contextId, shouldPlay)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
class PlaybackTargetPickerWasDismissed {
public:
    using Arguments = std::tuple<const WebCore::PlaybackTargetClientContextIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_PlaybackTargetPickerWasDismissed; }
    static const bool isSync = false;

    explicit PlaybackTargetPickerWasDismissed(const WebCore::PlaybackTargetClientContextIdentifier& contextId)
        : m_arguments(contextId)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(POINTER_LOCK)
class DidAcquirePointerLock {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidAcquirePointerLock; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(POINTER_LOCK)
class DidNotAcquirePointerLock {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidNotAcquirePointerLock; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(POINTER_LOCK)
class DidLosePointerLock {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidLosePointerLock; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class ClearWheelEventTestMonitor {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ClearWheelEventTestMonitor; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetShouldScaleViewToFitDocument {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetShouldScaleViewToFitDocument; }
    static const bool isSync = false;

    explicit SetShouldScaleViewToFitDocument(bool shouldScaleViewToFitDocument)
        : m_arguments(shouldScaleViewToFitDocument)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if ENABLE(VIDEO) && USE(GSTREAMER)
class DidEndRequestInstallMissingMediaPlugins {
public:
    using Arguments = std::tuple<uint32_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidEndRequestInstallMissingMediaPlugins; }
    static const bool isSync = false;

    explicit DidEndRequestInstallMissingMediaPlugins(uint32_t result)
        : m_arguments(result)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class SetUserInterfaceLayoutDirection {
public:
    using Arguments = std::tuple<uint32_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetUserInterfaceLayoutDirection; }
    static const bool isSync = false;

    explicit SetUserInterfaceLayoutDirection(uint32_t direction)
        : m_arguments(direction)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidGetLoadDecisionForIcon {
public:
    using Arguments = std::tuple<bool, const WebKit::CallbackID&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DidGetLoadDecisionForIcon; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(IPC::DataReference&&)>&&);
    static void cancelReply(CompletionHandler<void(IPC::DataReference&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_DidGetLoadDecisionForIconReply; }
    using AsyncReply = DidGetLoadDecisionForIconAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const IPC::SharedBufferDataReference& iconData);
    using Reply = std::tuple<IPC::SharedBufferDataReference&>;
    using ReplyArguments = std::tuple<IPC::DataReference>;
    DidGetLoadDecisionForIcon(bool decision, const WebKit::CallbackID& loadIdentifier)
        : m_arguments(decision, loadIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetUseIconLoadingClient {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetUseIconLoadingClient; }
    static const bool isSync = false;

    explicit SetUseIconLoadingClient(bool useIconLoadingClient)
        : m_arguments(useIconLoadingClient)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if ENABLE(GAMEPAD)
class GamepadActivity {
public:
    using Arguments = std::tuple<const Vector<WebKit::GamepadData>&, WebCore::EventMakesGamepadsVisible>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GamepadActivity; }
    static const bool isSync = false;

    GamepadActivity(const Vector<WebKit::GamepadData>& gamepadDatas, WebCore::EventMakesGamepadsVisible eventVisibility)
        : m_arguments(gamepadDatas, eventVisibility)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class FrameBecameRemote {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&, const WebCore::GlobalFrameIdentifier&, const WebCore::GlobalWindowIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_FrameBecameRemote; }
    static const bool isSync = false;

    FrameBecameRemote(const WebCore::FrameIdentifier& frameID, const WebCore::GlobalFrameIdentifier& remoteFrameIdentifier, const WebCore::GlobalWindowIdentifier& remoteWindowIdentifier)
        : m_arguments(frameID, remoteFrameIdentifier, remoteWindowIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RegisterURLSchemeHandler {
public:
    using Arguments = std::tuple<uint64_t, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RegisterURLSchemeHandler; }
    static const bool isSync = false;

    RegisterURLSchemeHandler(uint64_t identifier, const String& scheme)
        : m_arguments(identifier, scheme)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class URLSchemeTaskDidPerformRedirection {
public:
    using Arguments = std::tuple<uint64_t, uint64_t, const WebCore::ResourceResponse&, const WebCore::ResourceRequest&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_URLSchemeTaskDidPerformRedirection; }
    static const bool isSync = false;

    URLSchemeTaskDidPerformRedirection(uint64_t handlerIdentifier, uint64_t taskIdentifier, const WebCore::ResourceResponse& response, const WebCore::ResourceRequest& request)
        : m_arguments(handlerIdentifier, taskIdentifier, response, request)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class URLSchemeTaskDidReceiveResponse {
public:
    using Arguments = std::tuple<uint64_t, uint64_t, const WebCore::ResourceResponse&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_URLSchemeTaskDidReceiveResponse; }
    static const bool isSync = false;

    URLSchemeTaskDidReceiveResponse(uint64_t handlerIdentifier, uint64_t taskIdentifier, const WebCore::ResourceResponse& response)
        : m_arguments(handlerIdentifier, taskIdentifier, response)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class URLSchemeTaskDidReceiveData {
public:
    using Arguments = std::tuple<uint64_t, uint64_t, const IPC::SharedBufferDataReference&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_URLSchemeTaskDidReceiveData; }
    static const bool isSync = false;

    URLSchemeTaskDidReceiveData(uint64_t handlerIdentifier, uint64_t taskIdentifier, const IPC::SharedBufferDataReference& data)
        : m_arguments(handlerIdentifier, taskIdentifier, data)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class URLSchemeTaskDidComplete {
public:
    using Arguments = std::tuple<uint64_t, uint64_t, const WebCore::ResourceError&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_URLSchemeTaskDidComplete; }
    static const bool isSync = false;

    URLSchemeTaskDidComplete(uint64_t handlerIdentifier, uint64_t taskIdentifier, const WebCore::ResourceError& error)
        : m_arguments(handlerIdentifier, taskIdentifier, error)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetIsSuspended {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetIsSuspended; }
    static const bool isSync = false;

    explicit SetIsSuspended(bool suspended)
        : m_arguments(suspended)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if ENABLE(ATTACHMENT_ELEMENT)
class InsertAttachment {
public:
    using Arguments = std::tuple<const String&, const Optional<uint64_t>&, const String&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_InsertAttachment; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_InsertAttachmentReply; }
    using AsyncReply = InsertAttachmentAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    InsertAttachment(const String& identifier, const Optional<uint64_t>& fileSize, const String& fileName, const String& contentType)
        : m_arguments(identifier, fileSize, fileName, contentType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(ATTACHMENT_ELEMENT)
class UpdateAttachmentAttributes {
public:
    using Arguments = std::tuple<const String&, const Optional<uint64_t>&, const String&, const String&, const IPC::SharedBufferDataReference&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_UpdateAttachmentAttributes; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_UpdateAttachmentAttributesReply; }
    using AsyncReply = UpdateAttachmentAttributesAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    UpdateAttachmentAttributes(const String& identifier, const Optional<uint64_t>& fileSize, const String& contentType, const String& fileName, const IPC::SharedBufferDataReference& enclosingImageData)
        : m_arguments(identifier, fileSize, contentType, fileName, enclosingImageData)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(ATTACHMENT_ELEMENT)
class UpdateAttachmentIcon {
public:
    using Arguments = std::tuple<const String&, const WebKit::ShareableBitmap::Handle&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_UpdateAttachmentIcon; }
    static const bool isSync = false;

    UpdateAttachmentIcon(const String& identifier, const WebKit::ShareableBitmap::Handle& qlThumbnailHandle)
        : m_arguments(identifier, qlThumbnailHandle)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(APPLICATION_MANIFEST)
class GetApplicationManifest {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetApplicationManifest; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Optional<WebCore::ApplicationManifest>&&)>&&);
    static void cancelReply(CompletionHandler<void(Optional<WebCore::ApplicationManifest>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetApplicationManifestReply; }
    using AsyncReply = GetApplicationManifestAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Optional<WebCore::ApplicationManifest>& manifest);
    using Reply = std::tuple<Optional<WebCore::ApplicationManifest>&>;
    using ReplyArguments = std::tuple<Optional<WebCore::ApplicationManifest>>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class SetDefersLoading {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetDefersLoading; }
    static const bool isSync = false;

    explicit SetDefersLoading(bool defersLoading)
        : m_arguments(defersLoading)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if ENABLE(UI_PROCESS_PDF_HUD)
class ZoomPDFIn {
public:
    using Arguments = std::tuple<const WebKit::PDFPluginIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ZoomPDFIn; }
    static const bool isSync = false;

    explicit ZoomPDFIn(const WebKit::PDFPluginIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(UI_PROCESS_PDF_HUD)
class ZoomPDFOut {
public:
    using Arguments = std::tuple<const WebKit::PDFPluginIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ZoomPDFOut; }
    static const bool isSync = false;

    explicit ZoomPDFOut(const WebKit::PDFPluginIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(UI_PROCESS_PDF_HUD)
class SavePDF {
public:
    using Arguments = std::tuple<const WebKit::PDFPluginIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SavePDF; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(String&&, URL&&, IPC::DataReference&&)>&&);
    static void cancelReply(CompletionHandler<void(String&&, URL&&, IPC::DataReference&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_SavePDFReply; }
    using AsyncReply = SavePDFAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& filename, const URL& url, const IPC::DataReference& data);
    using Reply = std::tuple<String&, URL&, IPC::DataReference&>;
    using ReplyArguments = std::tuple<String, URL, IPC::DataReference>;
    explicit SavePDF(const WebKit::PDFPluginIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(UI_PROCESS_PDF_HUD)
class OpenPDFWithPreview {
public:
    using Arguments = std::tuple<const WebKit::PDFPluginIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_OpenPDFWithPreview; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(String&&, WebKit::FrameInfoData&&, IPC::DataReference&&, String&&)>&&);
    static void cancelReply(CompletionHandler<void(String&&, WebKit::FrameInfoData&&, IPC::DataReference&&, String&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_OpenPDFWithPreviewReply; }
    using AsyncReply = OpenPDFWithPreviewAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& filename, const WebKit::FrameInfoData& frameInfo, const IPC::DataReference& data, const String& uuid);
    using Reply = std::tuple<String&, WebKit::FrameInfoData&, IPC::DataReference&, String&>;
    using ReplyArguments = std::tuple<String, WebKit::FrameInfoData, IPC::DataReference, String>;
    explicit OpenPDFWithPreview(const WebKit::PDFPluginIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class UpdateCurrentModifierState {
public:
    using Arguments = std::tuple<const OptionSet<WebCore::PlatformEvent::Modifier>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_UpdateCurrentModifierState; }
    static const bool isSync = false;

    explicit UpdateCurrentModifierState(const OptionSet<WebCore::PlatformEvent::Modifier>& modifiers)
        : m_arguments(modifiers)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SimulateDeviceOrientationChange {
public:
    using Arguments = std::tuple<double, double, double>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SimulateDeviceOrientationChange; }
    static const bool isSync = false;

    SimulateDeviceOrientationChange(double alpha, double beta, double gamma)
        : m_arguments(alpha, beta, gamma)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if ENABLE(SPEECH_SYNTHESIS)
class SpeakingErrorOccurred {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SpeakingErrorOccurred; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(SPEECH_SYNTHESIS)
class BoundaryEventOccurred {
public:
    using Arguments = std::tuple<bool, const unsigned&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_BoundaryEventOccurred; }
    static const bool isSync = false;

    BoundaryEventOccurred(bool wordBoundary, const unsigned& charIndex)
        : m_arguments(wordBoundary, charIndex)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(SPEECH_SYNTHESIS)
class VoicesDidChange {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_VoicesDidChange; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class SetCanShowPlaceholder {
public:
    using Arguments = std::tuple<const WebCore::ElementContext&, bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetCanShowPlaceholder; }
    static const bool isSync = false;

    SetCanShowPlaceholder(const WebCore::ElementContext& context, bool canShowPlaceholder)
        : m_arguments(context, canShowPlaceholder)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if ENABLE(RESOURCE_LOAD_STATISTICS)
class WasLoadedWithDataTransferFromPrevalentResource {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_WasLoadedWithDataTransferFromPrevalentResource; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)
class ClearLoadedSubresourceDomains {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_ClearLoadedSubresourceDomains; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(RESOURCE_LOAD_STATISTICS)
class GetLoadedSubresourceDomains {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetLoadedSubresourceDomains; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(Vector<WebCore::RegistrableDomain>&&)>&&);
    static void cancelReply(CompletionHandler<void(Vector<WebCore::RegistrableDomain>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetLoadedSubresourceDomainsReply; }
    using AsyncReply = GetLoadedSubresourceDomainsAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const Vector<WebCore::RegistrableDomain>& domains);
    using Reply = std::tuple<Vector<WebCore::RegistrableDomain>&>;
    using ReplyArguments = std::tuple<Vector<WebCore::RegistrableDomain>>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if USE(SYSTEM_PREVIEW)
class SystemPreviewActionTriggered {
public:
    using Arguments = std::tuple<const WebCore::SystemPreviewInfo&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SystemPreviewActionTriggered; }
    static const bool isSync = false;

    SystemPreviewActionTriggered(const WebCore::SystemPreviewInfo& previewInfo, const String& message)
        : m_arguments(previewInfo, message)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(GTK) || PLATFORM(WPE)
class SendMessageToWebExtension {
public:
    using Arguments = std::tuple<const WebKit::UserMessage&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SendMessageToWebExtension; }
    static const bool isSync = false;

    explicit SendMessageToWebExtension(const WebKit::UserMessage& userMessage)
        : m_arguments(userMessage)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(GTK) || PLATFORM(WPE)
class SendMessageToWebExtensionWithReply {
public:
    using Arguments = std::tuple<const WebKit::UserMessage&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SendMessageToWebExtensionWithReply; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebKit::UserMessage&&)>&&);
    static void cancelReply(CompletionHandler<void(WebKit::UserMessage&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_SendMessageToWebExtensionWithReplyReply; }
    using AsyncReply = SendMessageToWebExtensionWithReplyAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const WebKit::UserMessage& replyMessage);
    using Reply = std::tuple<WebKit::UserMessage&>;
    using ReplyArguments = std::tuple<WebKit::UserMessage>;
    explicit SendMessageToWebExtensionWithReply(const WebKit::UserMessage& userMessage)
        : m_arguments(userMessage)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class StartTextManipulations {
public:
    using Arguments = std::tuple<const Vector<WebCore::TextManipulationController::ExclusionRule>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_StartTextManipulations; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void()>&&);
    static void cancelReply(CompletionHandler<void()>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_StartTextManipulationsReply; }
    using AsyncReply = StartTextManipulationsAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&);
    using Reply = std::tuple<>;
    using ReplyArguments = std::tuple<>;
    explicit StartTextManipulations(const Vector<WebCore::TextManipulationController::ExclusionRule>& exclusionRules)
        : m_arguments(exclusionRules)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CompleteTextManipulation {
public:
    using Arguments = std::tuple<const Vector<WebCore::TextManipulationController::ManipulationItem>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_CompleteTextManipulation; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&, Vector<WebCore::TextManipulationController::ManipulationFailure>&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&, Vector<WebCore::TextManipulationController::ManipulationFailure>&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_CompleteTextManipulationReply; }
    using AsyncReply = CompleteTextManipulationAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool allFailed, const Vector<WebCore::TextManipulationController::ManipulationFailure>& failures);
    using Reply = std::tuple<bool&, Vector<WebCore::TextManipulationController::ManipulationFailure>&>;
    using ReplyArguments = std::tuple<bool, Vector<WebCore::TextManipulationController::ManipulationFailure>>;
    explicit CompleteTextManipulation(const Vector<WebCore::TextManipulationController::ManipulationItem>& items)
        : m_arguments(items)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetOverriddenMediaType {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetOverriddenMediaType; }
    static const bool isSync = false;

    explicit SetOverriddenMediaType(const String& mediaType)
        : m_arguments(mediaType)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetProcessDisplayName {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_GetProcessDisplayName; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(String&&)>&&);
    static void cancelReply(CompletionHandler<void(String&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_GetProcessDisplayNameReply; }
    using AsyncReply = GetProcessDisplayNameAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, const String& displayName);
    using Reply = std::tuple<String&>;
    using ReplyArguments = std::tuple<String>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class UpdateCORSDisablingPatterns {
public:
    using Arguments = std::tuple<const Vector<String>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_UpdateCORSDisablingPatterns; }
    static const bool isSync = false;

    explicit UpdateCORSDisablingPatterns(const Vector<String>& patterns)
        : m_arguments(patterns)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetIsTakingSnapshotsForApplicationSuspension {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetIsTakingSnapshotsForApplicationSuspension; }
    static const bool isSync = false;

    explicit SetIsTakingSnapshotsForApplicationSuspension(bool isTakingSnapshotsForApplicationSuspension)
        : m_arguments(isTakingSnapshotsForApplicationSuspension)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetNeedsDOMWindowResizeEvent {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetNeedsDOMWindowResizeEvent; }
    static const bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetHasResourceLoadClient {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetHasResourceLoadClient; }
    static const bool isSync = false;

    explicit SetHasResourceLoadClient(bool has)
        : m_arguments(has)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SetCanUseCredentialStorage {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_SetCanUseCredentialStorage; }
    static const bool isSync = false;

    explicit SetCanUseCredentialStorage(bool canUse)
        : m_arguments(canUse)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if ENABLE(APP_HIGHLIGHTS)
class CreateAppHighlightInSelectedRange {
public:
    using Arguments = std::tuple<WebCore::CreateNewGroupForHighlight>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_CreateAppHighlightInSelectedRange; }
    static const bool isSync = false;

    explicit CreateAppHighlightInSelectedRange(WebCore::CreateNewGroupForHighlight createNewGroup)
        : m_arguments(createNewGroup)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(APP_HIGHLIGHTS)
class RestoreAppHighlights {
public:
    using Arguments = std::tuple<const Vector<WebKit::SharedMemory::IPCHandle>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_RestoreAppHighlights; }
    static const bool isSync = false;

    explicit RestoreAppHighlights(const Vector<WebKit::SharedMemory::IPCHandle>& memoryHandles)
        : m_arguments(memoryHandles)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class DispatchWheelEventWithoutScrolling {
public:
    using Arguments = std::tuple<const WebKit::WebWheelEvent&>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_DispatchWheelEventWithoutScrolling; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_DispatchWheelEventWithoutScrollingReply; }
    using AsyncReply = DispatchWheelEventWithoutScrollingAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool handled);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit DispatchWheelEventWithoutScrolling(const WebKit::WebWheelEvent& event)
        : m_arguments(event)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class LastNavigationWasAppBound {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebPage_LastNavigationWasAppBound; }
    static const bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebPage_LastNavigationWasAppBoundReply; }
    using AsyncReply = LastNavigationWasAppBoundAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(std::unique_ptr<IPC::Encoder>&&, IPC::Connection&, bool wasAppBound);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

} // namespace WebPage
} // namespace Messages
